import numpy as np
cimport numpy as np
cimport cython


DTYPE = np.int
ctypedef np.int_t DTYPE_INT_t
DTYPE_INTP = np.intp
ctypedef np.intp_t DTYPE_INTP_t
DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t


@cython.boundscheck(False)
def find_rows_containing_ID(np.ndarray[DTYPE_INT_t, ndim=2] input_array,
                            np.ndarray[DTYPE_INT_t, ndim=2] out):
    """
    Record the row in which ID appears in in_array, indexed by ID.

    Parameters
    ----------
    input_array : 2d array of ints
        The input array. The function will report in which rows it finds
        each ID.
    out : 2d array of ints, num_IDs x num_rows_in_input_array
        The output array. The number of rows sets the number of IDs to search
        for in input_array.
    """
    cdef int nrows = out.shape[0]
    cdef int i
    cdef np.ndarray[DTYPE_INT_t, ndim=2] contains_ID = np.empty_like(
        input_array)

    for i in range(nrows):
        contains_ID = np.equal(input_array, i).astype(int)
        out[i, :] = contains_ID.sum(axis=1)


@cython.boundscheck(False)
def create_patches_at_element(
        np.ndarray[DTYPE_INT_t, ndim=2] elements_at_patch,
        int number_of_elements, np.ndarray[DTYPE_INT_t, ndim=2] out):
    """
    """
    cdef int i
    cdef np.ndarray[DTYPE_INT_t, ndim=2] element_with_value = np.empty_like(
        (elements_at_patch), dtype=int)
    cdef np.ndarray[DTYPE_INTP_t, ndim=1] patches_with_element
    cdef int num_elements_here

    for i in range(number_of_elements):
        np.equal(elements_at_patch, i, out=element_with_value)
        patches_with_element = np.argwhere(element_with_value)[:, 0]
        num_elements_here = patches_with_element.size
        out[i, :num_elements_here] = patches_with_element


@cython.boundscheck(False)
def create_links_at_patch(np.ndarray[DTYPE_INT_t, ndim=2] nodes_at_patch,
                          np.ndarray[DTYPE_INT_t, ndim=2] links_at_node,
                          int number_of_patches,
                          np.ndarray[DTYPE_INT_t, ndim=2] out):

    cdef int i
    cdef np.ndarray[DTYPE_INT_t, ndim=1] nodes_on_patch = np.empty(
        nodes_at_patch.shape[1], dtype=int)
    cdef np.ndarray[DTYPE_INT_t, ndim=2] links_at_patch_nodes = np.empty(
        (nodes_at_patch.shape[1], links_at_node.shape[1]), dtype=int)
    cdef np.ndarray[DTYPE_INT_t, ndim=1] vals
    cdef np.ndarray[DTYPE_INTP_t, ndim=1] counts
    cdef np.ndarray[DTYPE_INT_t, ndim=1] duplicated_vals

    for i in range(number_of_patches):
        nodes_on_patch = nodes_at_patch[i, :]
        links_at_patch_nodes = links_at_node[nodes_on_patch, :]
        vals, counts = np.unique(links_at_patch_nodes,
                                 return_counts=True)
        duplicated_vals = vals[counts == 2]
        # if len==4, contains a -1; strip it
        out[i, :] = duplicated_vals[-3:]  # out of order, so far


@cython.boundscheck(False)
def _argsort_points_by_x_then_y(np.ndarray[DTYPE_INT_t, ndim=1] pts,
                                np.ndarray[DTYPE_INT_t, ndim=1] out):

    cdef np.ndarray[DTYPE_INT_t, ndim=1] a
    cdef np.ndarray[DTYPE_INT_t, ndim=1] b

    a = pts[:, 0].argsort(kind='mergesort')
    b = pts[a, 1].argsort(kind='mergesort')
    out[:] = a[b]


@cython.boundscheck(False)
def _anticlockwise_argsort_points(np.ndarray[DTYPE_INT_t, ndim=2] pts,
                                  np.ndarray[DTYPE_INT_t, ndim=1] out):

    cdef np.ndarray[DTYPE_FLOAT_t, ndim=1] midpt = pts.mean(axis=0)
    cdef np.ndarray[DTYPE_FLOAT_t, ndim=1] theta = np.empty(
        pts.shape[0], dtype=float)
    cdef double twopi = 2.*np.pi

    theta[:] = np.arctan2(pts[:, 1] - midpt[1], pts[:, 0] - midpt[0])
    theta[:] = theta % twopi
    out[:] = np.argsort(theta)




#! /usr/bin/env python
"""Functions to calculate flow distance."""
import numpy as np

from landlab import BAD_INDEX_VALUE, RasterModelGrid, FieldError


def calculate_flow__distance(grid, add_to_grid=False, noclobber=True):
    """
    Calculate the along flow distance from node to outlet.

    This utility calculates the along flow distance based on the results of
    running flow accumulation on the grid. It will use the connectivity
    used by the FlowAccumulator (e.g. D4, D8, Dinf).

    Parameters
    ----------
    grid : ModelGrid
    add_to_grid : boolean, optional
        Flag to indicate if the stream length field should be added to the
        grid. Default is False. The field name used is ``flow__distance``.
    noclobber : boolean, optional
        Flag to indicate if adding the field to the grid should not clobber an
        existing field with the same name. Default is True.

    Returns
    -------
    flow__distance : float ndarray
        The distance that has to be covered from an imaginary flow, located in
        each node of the grid, to reach the watershed's outlet.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator
    >>> from landlab.utils.flow__distance import calculate_flow__distance
    >>> mg = RasterModelGrid((5, 4), spacing=(1, 1))
    >>> elev = np.array([0.,  0.,  0., 0.,
    ...                  0., 21., 10., 0.,
    ...                  0., 31., 20., 0.,
    ...                  0., 32., 30., 0.,
    ...                  0.,  0.,  0., 0.])
    >>> _ = mg.add_field('node','topographic__elevation', elev)
    >>> mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
    ...                                        left_is_closed=True,
    ...                                        right_is_closed=True,
    ...                                        top_is_closed=True)
    >>> fr = FlowAccumulator(mg, flow_director = 'D8')
    >>> fr.run_one_step()
    >>> flow__distance = calculate_flow__distance(mg, add_to_grid=True, noclobber=False)
    >>> mg.at_node['flow__distance']
    array([ 0.        ,  0.        ,  0.        ,  0.        ,
            0.        ,  1.        ,  0.        ,  0.        ,
            0.        ,  1.41421356,  1.        ,  0.        ,
            0.        ,  2.41421356,  2.        ,  0.        ,
            0.        ,  0.        ,  0.        ,  0.        ])

    Now, let's change to D4 the flow_director method, which does not
    consider diagonal links bewtween nodes.

    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowAccumulator
    >>> from landlab.utils.flow__distance import calculate_flow__distance
    >>> mg = RasterModelGrid((5, 4), spacing=(1, 1))
    >>> elev = np.array([0.,  0.,  0., 0.,
    ...                  0., 21., 10., 0.,
    ...                  0., 31., 20., 0.,
    ...                  0., 32., 30., 0.,
    ...                  0.,  0.,  0., 0.])
    >>> _ = mg.add_field('node','topographic__elevation', elev)
    >>> mg.set_closed_boundaries_at_grid_edges(bottom_is_closed=True,
    ...                                        left_is_closed=True,
    ...                                        right_is_closed=True,
    ...                                        top_is_closed=True)
    >>> fr = FlowAccumulator(mg, flow_director = 'D4')
    >>> fr.run_one_step()
    >>> flow__distance = calculate_flow__distance(mg, add_to_grid=True,
    ...                                          noclobber=False)
    >>> mg.at_node['flow__distance']
    array([ 0.,  0.,  0.,  0.,
            0.,  1.,  0.,  0.,
            0.,  2.,  1.,  0.,
            0.,  3.,  2.,  0.,
            0.,  0.,  0.,  0.])

    The flow__distance utility can also work on irregular grids. For the example we
    will use a Hexagonal Model Grid, a special type of Voroni Grid that has
    regularly spaced hexagonal cells.

    >>> from landlab import HexModelGrid, FIXED_VALUE_BOUNDARY, CLOSED_BOUNDARY
    >>> from landlab.components import FlowAccumulator
    >>> from landlab.utils.flow__distance import calculate_flow__distance
    >>> dx = 1
    >>> hmg = HexModelGrid(5,3, dx)
    >>> _ = hmg.add_field('topographic__elevation',
    ...                   hmg.node_x + np.round(hmg.node_y),
    ...                   at = 'node')
    >>> hmg.status_at_node[hmg.boundary_nodes] = CLOSED_BOUNDARY
    >>> hmg.status_at_node[0] = FIXED_VALUE_BOUNDARY
    >>> fr = FlowAccumulator(hmg, flow_director = 'D4')
    >>> fr.run_one_step()
    >>> flow__distance = calculate_flow__distance(hmg,
    ...                                           add_to_grid=True,
    ...                                           noclobber=False)
    >>> hmg.at_node['flow__distance']
    array([ 0.,  0.,  0.,
            0.,  1.,  2.,  0.,
            0.,  2.,  2.,  3.,  0.,
            0.,  3.,  3.,  0.,
            0.,  0.,  0.])
    """
    # check that flow__receiver nodes exists
    if 'flow__receiver_node' not in grid.at_node:
        raise FieldError("A 'flow__receiver_node' field is required at the "
                         "nodes of the input grid.")
    if 'flow__upstream_node_order' not in grid.at_node:
        raise FieldError("A 'flow__upstream_node_order' field is required at the "
                         "nodes of the input grid.")

    # get the reciever nodes, depending on if this is to-one, or to-multiple,
    # we'll need to get a different at-node field.
    if 'flow__receiver_nodes' in grid.at_node:
        to_one = False
        flow__receiver_node = grid.at_node['flow__receiver_nodes']
    else:
        to_one = True
        flow__receiver_node = grid.at_node['flow__receiver_node']

    # get the upstream node order
    flow__upstream_node_order = grid.at_node['flow__upstream_node_order']

    # get downstream flow link lengths, result depends on type of grid.
    if isinstance(grid, RasterModelGrid):
        if to_one:
            flow_link_lengths = grid.length_of_d8[grid.at_node['flow__link_to_receiver_node']]
        else:
            flow_link_lengths = grid.length_of_d8[grid.at_node['flow__links_to_receiver_nodes']]
    else:
        if to_one:
            flow_link_lengths = grid.length_of_link[grid.at_node['flow__link_to_receiver_node']]
        else:
            flow_link_lengths = grid.length_of_link[grid.at_node['flow__links_to_receiver_nodes']]

    # create an array that representes the outlet lengths.
    flow__distance = np.zeros(grid.nodes.size)

    # iterate through the flow__upstream_node_order, this will already have
    # identified the locations of the outlet nodes and have
    for node in flow__upstream_node_order:

        # get flow recievers
        reciever = flow__receiver_node[node]

        # assess if this is a to one (D8/D4) or to multiple (Dinf, MFD)
        # flow directing method.
        if to_one:
            potential_outlet = reciever
        else:
            # if this is an outlet, the first element of the recievers will be
            # the nodes ID.
            potential_outlet = reciever[0]

        # assess if this is an outlet or not.
        if potential_outlet == node:
            not_outlet = False
        else:
            not_outlet = True

        # if not an outlet
        if not_outlet:

            # deal with the two cases of route to one and route to multiple.
            if to_one:
                # get the stream length of the downstream node
                downstream_stream_length = flow__distance[reciever]

                # get the stream segment length from this node to its downstream
                # neigbor
                stream_increment_length = flow_link_lengths[node]

            else:
                # non-existant links are coded with -1
                useable_recievers = np.where(reciever != BAD_INDEX_VALUE)[0]

                # we will have the stream flow to the downstream node with the
                # shortest distance to the outlet.
                # in the event of a tie, we will choose the shorter link length.

                # get the flow distances of the downstream nodes
                potential_downstream_stream_lengths = flow__distance[flow__receiver_node[node]][useable_recievers]

                # get the stream segment lengths from this node to its downstream
                # neighbor
                potential_stream_increment_lengths = flow_link_lengths[node][useable_recievers]

                # get the lowest downstream stream length.
                downstream_stream_length = np.min(potential_downstream_stream_lengths)

                # determine which of the stream increments flowed to this
                # downstream neighbor.
                which_link = np.where(potential_downstream_stream_lengths == downstream_stream_length)[0]

                # and choose the smallest of these links.
                stream_increment_length = np.min(potential_stream_increment_lengths[which_link])

            # set the total stream length of this node
            flow__distance[node] = downstream_stream_length + stream_increment_length

    # store on the grid
    if add_to_grid:
        grid.add_field('node', 'flow__distance', flow__distance, noclobber=noclobber)

    return flow__distance



import numpy as np
cimport numpy as np
cimport cython


from libc.string cimport memcpy


ctypedef np.int_t INT_t
ctypedef np.float_t FLOAT_t


cdef _pad_jaggedarray(void * data, long * offset, size_t n_rows, size_t size,
                      void * buff, size_t n_cols):
    cdef int row
    cdef void * dst = buff
    cdef void * src = data

    for row in range(n_rows):
        dst = <char *>buff + n_cols * row * size
        src = <char *>data + offset[row] * size

        memcpy(dst, src, size * (offset[row + 1] - offset[row]))


@cython.boundscheck(False)
def unravel(np.ndarray data,
            np.ndarray[long, ndim=1, mode="c"] offset,
            np.ndarray out):
    _pad_jaggedarray(data.data, &offset[0], len(offset) - 1,
                     data.itemsize, out.data, out.shape[1])



import numpy as np
from scipy.spatial import Voronoi

from ..graph import Graph
from ..dual import DualGraph
from .voronoi import VoronoiGraph
from ...utils.jaggedarray import JaggedArray
from .voronoi_helpers import VoronoiConverter


def ugrid_from_voronoi_dual(node_y_and_x, min_cell_size=3,
                            max_node_spacing=None):
    from .voronoi import ugrid_from_voronoi
    from ..ugrid import (ugrid_from_unstructured,
                         update_node_at_cell, update_nodes_at_face)

    voronoi = Voronoi(list(zip(node_y_and_x[1], node_y_and_x[0])))

    converter = VoronoiConverter(voronoi, min_patch_size=min_cell_size)

    corners = converter.get_nodes()
    corners = (corners[:, 1], corners[:, 0])
    faces = converter.get_nodes_at_link()
    cells = converter.get_links_at_patch()
    cells = [cell for cell in JaggedArray(cells)]

    node_at_cell = converter.get_corner_at_patch()
    nodes_at_face = converter.get_corners_at_link()

    dual = ugrid_from_unstructured(corners, faces, cells)

    return dual, node_at_cell, nodes_at_face


class DualVoronoiGraph(VoronoiGraph, DualGraph):

    """Dual graph of a voronoi grid."""

    def __init__(self, node_y_and_x, min_cell_size=3, **kwds):
        """Create a voronoi grid.

        Parameters
        ----------
        nodes : tuple of array_like
            Coordinates of every node. First *y*, then *x*.

        Examples
        --------
        >>> from landlab.graph import DualVoronoiGraph
        >>> node_x = [0, 1, 2, 3,
        ...           0.2, 1.2, 2.2, 3.2,
        ...           0.4, 1.4, 2.4, 3.4]
        >>> node_y = [0, 0, 0, 0,
        ...           1, 1, 1, 1,
        ...           2, 2, 2, 2]
        >>> graph = DualVoronoiGraph((node_y, node_x))
        >>> graph.x_of_corner
        array([ 1.5,  2.5,  0.7,  1.7,  2.7,  0.7,  1.7,  2.7,  0.9,  1.9])
        >>> graph.y_of_corner # doctest: +NORMALIZE_WHITESPACE
        array([ 0.42,  0.42,  0.58,  0.58,  0.58,  1.42,  1.42,  1.42,  1.58,
                1.58])
        >>> graph.corners_at_face # doctest: +NORMALIZE_WHITESPACE
        array([[2, 0], [0, 3], [3, 1], [1, 4],
               [2, 5], [3, 6], [4, 7],
               [5, 8], [8, 6], [6, 9], [9, 7]])
        >>> graph.faces_at_corner # doctest: +NORMALIZE_WHITESPACE
        array([[ 1,  0, -1], [ 3,  2, -1],
               [ 4,  0, -1], [ 5,  1,  2], [ 6,  3, -1],
               [ 7,  4, -1], [ 9,  8,  5], [10,  6, -1],
               [ 7,  8, -1], [ 9, 10, -1]])
        >>> graph.node_at_cell
        array([5, 6])
        """
        max_node_spacing = kwds.pop('max_node_spacing', None)
        (dual,
         node_at_cell,
         nodes_at_face) = ugrid_from_voronoi_dual(node_y_and_x,
                                                  min_cell_size=min_cell_size,
                                                  max_node_spacing=max_node_spacing)

        self._dual = Graph(dual, sort=False)
        VoronoiGraph.__init__(self, node_y_and_x,
                              max_node_spacing=max_node_spacing, sort=False)
        DualGraph.__init__(self, node_at_cell=node_at_cell,
                           nodes_at_face=nodes_at_face, sort=True)



from numpy.testing import assert_array_equal, assert_array_almost_equal
import numpy as np

from landlab.graph.sort.sort import remap


def test_remap():
    src = np.array([1, 2, 3, 4])
    mapping = np.array([-1, 10, 20, 30, 40])

    rtn = remap(src, mapping)
    assert_array_equal(rtn, [10, 20, 30, 40])
    assert rtn is not src


def test_remap_inplace():
    src = np.array([1, 2, 3, 4])
    mapping = np.array([-1, 10, 20, 30, 40])

    rtn = remap(src, mapping, inplace=True)

    assert_array_equal(rtn, [10, 20, 30, 40])
    assert rtn is src


def test_remap_out():
    src = np.array([1, 2, 3, 4])
    dst = np.empty_like(src)
    mapping = np.array([-1, 10, 20, 30, 40])

    rtn = remap(src, mapping, out=dst)

    assert_array_equal(rtn, [10, 20, 30, 40])
    assert rtn is dst



import numpy as np
from six import string_types

from landlab import Component
from landlab import RasterModelGrid
from landlab.utils.return_array import return_array_at_node
from .cfuncs import calculate_qs_in
DEFAULT_MINIMUM_TIME_STEP = 0.001 # default minimum time step duration

class _GeneralizedErosionDeposition(Component):
    """ Base class for erosion-deposition type components.

    More documenation here.
    """

    _name= 'ErosionDeposition'

    _input_var_names = (
        'flow__receiver_node',
        'flow__upstream_node_order',
        'topographic__steepest_slope',
        'drainage_area',
    )

    _output_var_names = (
        'topographic__elevation'
    )

    _var_units = {
        'flow__receiver_node': '-',
        'flow__upstream_node_order': '-',
        'topographic__steepest_slope': '-',
        'drainage_area': 'm**2',
        'topographic__elevation': 'm',
    }

    _var_mapping = {
        'flow__receiver_node': 'node',
        'flow__upstream_node_order': 'node',
        'topographic__steepest_slope': 'node',
        'drainage_area': 'node',
        'topographic__elevation': 'node',
    }

    _var_doc = {
        'flow__receiver_node':
            'Node array of receivers (node that receives flow from current '
            'node)',
        'flow__upstream_node_order':
            'Node array containing downstream-to-upstream ordered list of '
            'node IDs',
        'topographic__steepest_slope':
            'Topographic slope at each node',
        'drainage_area':
            "Upstream accumulated surface area contributing to the node's "
            "discharge",
        'topographic__elevation':
            'Land surface topographic elevation',
    }

    def __init__(self, grid, m_sp=None, n_sp=None,
                 phi=None, F_f=None, v_s=None,
                 discharge_field='surface_water__discharge',
                 dt_min=DEFAULT_MINIMUM_TIME_STEP):
        """Initialize the ErosionDeposition model.

        Parameters
        ----------
        grid : ModelGrid
            Landlab ModelGrid object
        m_sp : float
            Drainage area exponent (units vary)
        n_sp : float
            Slope exponent (units vary)
        phi : float
            Sediment porosity [-].
        F_f : float
            Fraction of eroded material that turns into "fines" that do not
            contribute to (coarse) sediment load. Defaults to zero.
        v_s : float
            Effective settling velocity for chosen grain size metric [L/T].
        discharge_field : float, field name, or array
            Discharge [L^2/T].
        dt_min : float, optional
            Only applies when adaptive solver is used. Minimum timestep that
            adaptive solver will use when subdividing unstable timesteps.
            Default values is 0.001. [T].
        """
        super(_GeneralizedErosionDeposition, self).__init__(grid)

        self.flow_receivers = grid.at_node['flow__receiver_node']
        self.stack = grid.at_node['flow__upstream_node_order']
        self.topographic__elevation = grid.at_node['topographic__elevation']
        self.slope = grid.at_node['topographic__steepest_slope']
        self.link_to_reciever = grid.at_node['flow__link_to_receiver_node']
        self.cell_area_at_node = grid.cell_area_at_node

        if isinstance(grid, RasterModelGrid):
            self.link_lengths = grid.length_of_d8
        else:
            self.link_lengths = grid.length_of_link

        if 'sediment__flux' in grid.at_node:
            self.qs = grid.at_node['sediment__flux']
        else:
            self.qs = grid.add_zeros(
                'sediment__flux', at='node', dtype=float)

        self.q = return_array_at_node(grid, discharge_field)

        # Create arrays for sediment influx at each node, discharge to the
        # power "m", and deposition rate
        self.qs_in = np.zeros(grid.number_of_nodes)
        self.Q_to_the_m = np.zeros(grid.number_of_nodes)
        self.S_to_the_n = np.zeros(grid.number_of_nodes)
        self.depo_rate = np.zeros(grid.number_of_nodes)

        # store other constants
        self.m_sp = float(m_sp)
        self.n_sp = float(n_sp)
        self.phi = float(phi)
        self.v_s = float(v_s)
        self.dt_min = dt_min
        self.F_f = float(F_f)

        if phi >= 1.0:
            raise ValueError('Porosity must be < 1.0')

        if F_f > 1.0:
            raise ValueError('Fraction of fines must be <= 1.0')

        if phi < 0.0:
            raise ValueError('Porosity must be > 0.0')

        if F_f < 0.0:
            raise ValueError('Fraction of fines must be > 0.0')

    def _update_flow_link_slopes(self):
        """Updates gradient between each core node and its receiver.

        Used to update slope values between sub-time-steps, when we do not
        re-run flow routing.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> from landlab.components import FlowAccumulator
        >>> from landlab.components.erosion_deposition.generalized_erosion_deposition import _GeneralizedErosionDeposition
        >>> rg = RasterModelGrid((3, 4))
        >>> z = rg.add_zeros('node', 'topographic__elevation')
        >>> z[:] = rg.x_of_node + rg.y_of_node
        >>> fa = FlowAccumulator(rg, flow_director='FlowDirectorD8')
        >>> fa.run_one_step()
        >>> rg.at_node['topographic__steepest_slope'][5:7]
        array([ 1.41421356,  1.41421356])
        >>> sp = _GeneralizedErosionDeposition(rg, phi=0.1, v_s=0.001,
        ...                                    m_sp=0.5, n_sp=1.0, F_f=0)
        >>> z *= 0.1
        >>> sp._update_flow_link_slopes()
        >>> rg.at_node['topographic__steepest_slope'][5:7]
        array([ 0.14142136,  0.14142136])
        """
        self.slope[:] = ((self.topographic__elevation -
                          self.topographic__elevation[self.flow_receivers]) /
                         self.link_lengths[self.link_to_reciever])

    def _calc_hydrology(self):
        self.Q_to_the_m[:] = np.power(self.q, self.m_sp)



"""Define a graph of nodes-links-patches.

Examples
--------

>>> from landlab.graph import Graph

>>> node_x, node_y = [0, 0, 0, 1, 1, 1, 2, 2, 2], [0, 1, 2, 0, 1, 2, 0, 1, 2]
>>> graph = Graph((node_y, node_x))
>>> graph.x_of_node
array([ 0.,  1.,  2.,  0.,  1.,  2.,  0.,  1.,  2.])
>>> graph.y_of_node
array([ 0.,  0.,  0.,  1.,  1.,  1.,  2.,  2.,  2.])

>>> links = ((0, 1), (1, 2),
...          (0, 3), (1, 4), (2, 5),
...          (3, 4), (4, 5),
...          (3, 6), (4, 7), (5, 8),
...          (6, 7), (7, 8))
>>> graph = Graph((node_y, node_x), links=links)
>>> graph.nodes_at_link # doctest: +NORMALIZE_WHITESPACE
array([[0, 1], [1, 2],
       [0, 3], [1, 4], [2, 5],
       [3, 4], [4, 5],
       [3, 6], [4, 7], [5, 8],
       [6, 7], [7, 8]])
>>> graph.node_at_link_head
array([1, 2, 3, 4, 5, 4, 5, 6, 7, 8, 7, 8])
>>> graph.node_at_link_tail
array([0, 1, 0, 1, 2, 3, 4, 3, 4, 5, 6, 7])

>>> graph.links_at_node # doctest: +NORMALIZE_WHITESPACE
array([[ 0,  2, -1, -1], [ 1,  3,  0, -1], [ 4,  1, -1, -1],
       [ 5,  7,  2, -1], [ 6,  8,  5,  3], [ 9,  6,  4, -1],
       [10,  7, -1, -1], [11, 10,  8, -1], [11,  9, -1, -1]])

>>> graph.link_dirs_at_node # doctest: +NORMALIZE_WHITESPACE
array([[-1, -1,  0,  0], [-1, -1,  1,  0], [-1,  1,  0,  0],
       [-1, -1,  1,  0], [-1, -1,  1,  1], [-1,  1,  1,  0],
       [-1,  1,  0,  0], [-1,  1,  1,  0], [ 1,  1,  0,  0]])

>>> patches = ((5, 3, 0, 2), (6, 4, 1, 3), (10, 8, 5, 7), (11, 9, 6, 8))
>>> graph = Graph((node_y, node_x), links=links, patches=patches)
>>> graph.links_at_patch
array([[ 3,  5,  2,  0],
       [ 4,  6,  3,  1],
       [ 8, 10,  7,  5],
       [ 9, 11,  8,  6]])
>>> graph.nodes_at_patch
array([[4, 3, 0, 1],
       [5, 4, 1, 2],
       [7, 6, 3, 4],
       [8, 7, 4, 5]])
"""
import six
from six.moves import range

import numpy as np
import xarray as xr
import json

from ..core.utils import as_id_array, argsort_points_by_x_then_y
from ..utils.jaggedarray import flatten_jagged_array
from ..utils.decorators import store_result_in_grid, read_only_array
from .sort import sort_graph, reindex_by_xy, reorder_links_at_patch
from .object.at_node import get_links_at_node
from .object.at_patch import get_nodes_at_patch
from .quantity.of_link import (get_angle_of_link, get_length_of_link,
                               get_midpoint_of_link)
from .quantity.of_patch import get_centroid_of_patch, get_area_of_patch

from .sort.sort import reverse_one_to_many, reorient_link_dirs

from .ugrid import update_nodes_at_patch, ugrid_from_unstructured


def _parse_sorting_opt(sorting):
    SORTING_OPTS = ('xy', 'ccw', 'ne')

    as_dict = None

    if isinstance(sorting, bool):
        as_dict = dict([(opt, format) for opt in SORTING_OPTS])
    elif isinstance(sorting, dict):
        as_dict = dict(sorting.items())
        for opt in SORTING_OPTS:
            sorting.setdefault(opt, True)

    return as_dict


def find_perimeter_nodes(graph):
    """Find nodes on the perimeter of a graph.

    Uses a convex hull to locate the perimeter nodes of a graph.

    Parameters
    ----------
    graph : graph_like
        A Graph of nodes (just requires *xy_of_node*).

    Returns
    -------
    ndarray of int
        Identifiers of the perimeter nodes.
    """
    from scipy.spatial import ConvexHull

    hull = ConvexHull(graph.xy_of_node, qhull_options='Qt')
    return as_id_array(hull.vertices)


class thawed(object):
    def __init__(self, graph):
        self._graph = graph
        self._initially_frozen = graph.frozen

    def __enter__(self):
        self._graph.thaw()

    def __exit__(self, ex_type, ex_value, traceback):
        if self._initially_frozen:
            self._graph.freeze()


class Graph(object):

    """Define the connectivity of a graph of nodes, links, and patches."""

    def __init__(self, mesh, **kwds):
        """Define a graph of connected nodes.

        Parameters
        ----------
        mesh : Dataset
            xarray Dataset that defines the topology in ugrid format.
        """
        if not isinstance(mesh, xr.Dataset):
            node_y_and_x = mesh
            links = kwds.get('links', None)
            patches = kwds.get('patches', None)
            mesh = ugrid_from_unstructured(node_y_and_x, links=links,
                                           patches=patches)
        self._ds = mesh

        self._frozen = False
        self.freeze()

        if kwds.get('sort', True):
            Graph.sort(self)

        self._origin = (0., 0.)

    @property
    def frozen(self):
        return self._frozen

    def thawed(self):
        return thawed(self)

    def sort(self):
        with self.thawed():
            reorient_link_dirs(self)
            sorted_nodes, sorted_links, sorted_patches = reindex_by_xy(self)
            reorder_links_at_patch(self)

        return sorted_nodes, sorted_links, sorted_patches

    def freeze(self):
        """Freeze the graph by making arrays read-only."""
        for var in self.ds.variables:
            self.ds[var].values.flags.writeable = False
        self._frozen = True

    def thaw(self):
        """Thaw the graph by making arrays writable."""
        for var in self.ds.variables:
            self.ds[var].values.flags.writeable = True
        self._frozen = False

    def _add_variable(self, name, var, dims=None, attrs=None):
        kwds = dict(data=var, dims=dims, attrs=attrs)
        self.ds.update({name: xr.DataArray(**kwds)})
        if self._frozen:
            self.freeze()

    @property
    def ds(self):
        return self._ds

    def to_dict(self):
        return self.ds.to_dict()

    def to_json(self):
        return json.dumps(self.ds.to_dict())

    def to_netcdf(self, *args, **kwds):
        """Write graph contents to a netCDF file.

        See xarray.Dataset.to_netcdf for a complete list of parameters.
        Below are only the most common.

        Parameters
        ----------
        path : str, optional
            Path to which to save this graph.
        mode : {'w', 'a'}, optional
            Write ('w') or append ('a') mode. If mode='w', any
            existing file at this location will be overwritten.
        format : {'NETCDF4', 'NETCDF4_CLASSIC', 'NETCDF3_64BIT', 'NETCDF3_CLASSIC'}, optional
            File format for the resulting netCDF file:

            * NETCDF4: Data is stored in an HDF5 file, using netCDF4 API
              features.
            * NETCDF4_CLASSIC: Data is stored in an HDF5 file, using only
              netCDF 3 compatible API features.
            * NETCDF3_64BIT: 64-bit offset version of the netCDF 3 file format,
              which fully supports 2+ GB files, but is only compatible with
              clients linked against netCDF version 3.6.0 or later.
            * NETCDF3_CLASSIC: The classic netCDF 3 file format. It does not
              handle 2+ GB files very well.

            All formats are supported by the netCDF4-python library.
            scipy.io.netcdf only supports the last two formats.

            The default format is NETCDF4 if you are saving a file to disk and
            have the netCDF4-python library available. Otherwise, xarray falls
            back to using scipy to write netCDF files and defaults to the
            NETCDF3_64BIT format (scipy does not support netCDF4).
        """
        self.ds.to_netcdf(*args, **kwds)

    @classmethod
    def from_netcdf(cls, fname):
        return cls.from_dataset(xr.open_dataset(fname))

    @classmethod
    def from_dict(cls, meta):
        return cls((meta['y_of_node'], meta['x_of_node']),
                   links=meta.get('nodes_at_link', None),
                   patches=meta.get('links_at_patch', None))

    @classmethod
    def load(cls, source):
        if isinstance(source, six.string_types):
            return cls.from_netcdf(source)
        elif isinstance(source, (dict, xr.Dataset)):
            return cls.from_dict(source)
        else:
            raise ValueError('source must be dict-like or NetCDF ({type})'.format(type=type(source)))

    def __str__(self):
        return str(self.ds)

    def __repr__(self):
        return repr(self.ds)

    @property
    def ndim(self):
        return 2

    @property
    @store_result_in_grid()
    @read_only_array
    def xy_of_node(self):
        """Get x and y-coordinates of node.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1]
        >>> graph = Graph((node_y, node_x))
        >>> graph.xy_of_node[:, 0]
        array([ 0.,  1.,  2.,  0.,  1.,  2.])
        >>> graph.xy_of_node[:, 1]
        array([ 0.,  0.,  0.,  1.,  1.,  1.])
        """
        return np.stack((self.x_of_node, self.y_of_node)).T.copy()

    @property
    def x_of_node(self):
        """Get x-coordinate of node.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1]
        >>> graph = Graph((node_y, node_x))
        >>> graph.x_of_node
        array([ 0.,  1.,  2.,  0.,  1.,  2.])
        """
        return self.ds['x_of_node'].values

    @property
    def y_of_node(self):
        """Get y-coordinate of node.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1]
        >>> graph = Graph((node_y, node_x))
        >>> graph.y_of_node
        array([ 0.,  0.,  0.,  1.,  1.,  1.])
        """
        return self.ds['y_of_node'].values

    @property
    def nodes(self):
        """Get identifier for each node.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1]
        >>> graph = Graph((node_y, node_x))
        >>> graph.nodes
        array([0, 1, 2, 3, 4, 5])
        """
        return self.ds['node'].values

    @property
    @store_result_in_grid()
    @read_only_array
    def perimeter_nodes(self):
        return find_perimeter_nodes(self)

    @property
    def number_of_nodes(self):
        """Get total number of nodes.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1]
        >>> graph = Graph((node_y, node_x))
        >>> graph.number_of_nodes == 6
        True
        """
        return self.ds.dims['node']

    @property
    def nodes_at_link(self):
        """Get nodes at either end of links.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1, 2, 2, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.nodes_at_link # doctest: +NORMALIZE_WHITESPACE
        array([[0, 1], [1, 2],
               [0, 3], [1, 4], [2, 5],
               [3, 4], [4, 5],
               [3, 6], [4, 7], [5, 8],
               [6, 7], [7, 8]])
        """
        return self.ds['nodes_at_link'].values

    @property
    def node_at_link_tail(self):
        """Get nodes at link tail.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1, 2, 2, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.node_at_link_tail
        array([0, 1, 0, 1, 2, 3, 4, 3, 4, 5, 6, 7])
        """
        return self.nodes_at_link[:, 0]

    @property
    def node_at_link_head(self):
        """Get nodes at link head.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1, 2, 2, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.node_at_link_head
        array([1, 2, 3, 4, 5, 4, 5, 6, 7, 8, 7, 8])
        """
        return self.nodes_at_link[:, 1]

    @property
    def number_of_links(self):
        """Get nodes at link head.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1, 2, 2, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.number_of_links == 12
        True
        """
        try:
            return self.ds.dims['link']
        except KeyError:
            return 0

    @property
    def links_at_patch(self):
        """Get the links that define a patch.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1, 2, 2, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> patches = ((0, 3, 5, 2), (1, 4, 6, 3))
        >>> graph = Graph((node_y, node_x), links=links, patches=patches)
        >>> graph.links_at_patch
        array([[3, 5, 2, 0],
               [4, 6, 3, 1]])
        """
        return self.ds['links_at_patch'].values

    @property
    # @store_result_in_grid()
    @read_only_array
    def nodes_at_patch(self):
        """Get the nodes that define a patch.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = ([0, 1, 2, 0, 1, 2, 0, 1, 2],
        ...                   [0, 0, 0, 1, 1, 1, 2, 2, 2])
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> patches = ((0, 3, 5, 2), (1, 4, 6, 3))
        >>> graph = Graph((node_y, node_x), links=links, patches=patches)
        >>> graph.nodes_at_patch
        array([[4, 3, 0, 1],
               [5, 4, 1, 2]])
        """
        return get_nodes_at_patch(self)

    @property
    @store_result_in_grid()
    @read_only_array
    def patches_at_node(self):
        """Get the patches that touch each node.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = ([0, 1, 2, 0, 1, 2],
        ...                   [0, 0, 0, 1, 1, 1])
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5))
        >>> patches = ((0, 3, 5, 2), (1, 4, 6, 3))
        >>> graph = Graph((node_y, node_x), links=links, patches=patches)
        >>> graph.patches_at_node # doctest: +NORMALIZE_WHITESPACE
        array([[ 0, -1], [ 0,  1], [ 1, -1],
               [ 0, -1], [ 0,  1], [ 1, -1]])
        """
        return reverse_one_to_many(self.nodes_at_patch)

    @property
    @store_result_in_grid()
    @read_only_array
    def patches_at_link(self):
        """Get the patches on either side of each link.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = ([0, 1, 2, 0, 1, 2],
        ...                   [0, 0, 0, 1, 1, 1])
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5))
        >>> patches = ((0, 3, 5, 2), (1, 4, 6, 3))
        >>> graph = Graph((node_y, node_x), links=links, patches=patches)
        >>> graph.patches_at_link # doctest: +NORMALIZE_WHITESPACE
        array([[ 0, -1], [ 1, -1],
               [ 0, -1], [ 0,  1], [ 1, -1],
               [ 0, -1], [ 1, -1]])
        """
        return reverse_one_to_many(self.links_at_patch, min_counts=2)
        try:
            return self.ds['patches_at_link'].values
        except KeyError:
            patches_at_link = xr.DataArray(
                data=reverse_one_to_many(self.links_at_patch, min_counts=2),
                dims=('link', 'Two'),
                attrs={'cf_role': 'edge_node_connectivity',
                       'long_name': 'patches on either side of a link',
                       'start_index': 0})
            self.ds.update({'patches_at_link': patches_at_link})
            return self.ds['patches_at_link'].values

    @property
    def number_of_patches(self):
        """Get the number of patches.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1, 2, 2, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> patches = ((0, 3, 5, 2), (1, 4, 6, 3))
        >>> graph = Graph((node_y, node_x), links=links, patches=patches)
        >>> graph.number_of_patches == 2
        True
        """
        try:
            return self.ds.dims['patch']
        except KeyError:
            return 0

    @property
    def links_at_node(self):
        """Get links touching a node.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x = [0, 1, 2, 0, 1, 2, 0, 1, 2]
        >>> node_y = [0, 0, 0, 1, 1, 1, 2, 2, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.links_at_node # doctest: +NORMALIZE_WHITESPACE
        array([[ 0,  2, -1, -1], [ 1,  3,  0, -1], [ 4,  1, -1, -1],
               [ 5,  7,  2, -1], [ 6,  8,  5,  3], [ 9,  6,  4, -1],
               [10,  7, -1, -1], [11, 10,  8, -1], [11,  9, -1, -1]])
        """
        try:
            return self._links_at_node
        except AttributeError:
            (self._links_at_node,
             self._link_dirs_at_node) = self._create_links_and_dirs_at_node()
            return self._links_at_node

    def _create_links_and_dirs_at_node(self):
        return get_links_at_node(self, sort=True)
        # return get_links_at_node(self, sort=self._sorting['ccw'])

    @property
    def link_dirs_at_node(self):
        """Get directions of links touching a node.

        Examples
        --------
        >>> from landlab.graph import Graph
        >>> node_x, node_y = [0, 1, 2, 0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1, 2, 2, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.link_dirs_at_node # doctest: +NORMALIZE_WHITESPACE
        array([[-1, -1,  0,  0], [-1, -1,  1,  0], [-1,  1,  0,  0],
               [-1, -1,  1,  0], [-1, -1,  1,  1], [-1,  1,  1,  0],
               [-1,  1,  0,  0], [-1,  1,  1,  0], [ 1,  1,  0,  0]])
        """
        try:
            return self._link_dirs_at_node
        except AttributeError:
            (self._links_at_node,
             self._link_dirs_at_node) = self._create_links_and_dirs_at_node()
            return self._link_dirs_at_node

    @property
    @store_result_in_grid()
    @read_only_array
    def angle_of_link(self):
        """Get the angle of each link.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph import Graph

        >>> node_x, node_y = ([0, 1, 2, 0, 1, 2],
        ...                   [0, 0, 0, 1, 1, 1])
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.angle_of_link * 180. / np.pi
        array([  0.,   0.,  90.,  90.,  90.,   0.,   0.])
        """
        return get_angle_of_link(self)

    @property
    @store_result_in_grid()
    @read_only_array
    def length_of_link(self):
        """Get the length of links.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph import UniformRectilinearGraph

        >>> graph = UniformRectilinearGraph((2, 3), spacing=(1, 2))
        >>> graph.length_of_link
        array([ 2.,  2.,  1.,  1.,  1.,  2.,  2.])
        """
        return get_length_of_link(self)

    @property
    @store_result_in_grid()
    @read_only_array
    def midpoint_of_link(self):
        """Get the middle of links.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.graph import UniformRectilinearGraph

        >>> graph = UniformRectilinearGraph((2, 3), spacing=(1, 2))
        >>> graph.midpoint_of_link # doctest: +NORMALIZE_WHITESPACE
        array([[ 1. ,  0. ], [ 3. ,  0. ],
               [ 0. ,  0.5], [ 2. ,  0.5], [ 4. ,  0.5],
               [ 1. ,  1. ], [ 3. ,  1. ]])
        """
        return get_midpoint_of_link(self)

    @property
    @store_result_in_grid()
    @read_only_array
    def xy_of_link(self):
        return get_midpoint_of_link(self)

    @property
    @store_result_in_grid()
    @read_only_array
    def xy_of_patch(self):
        return get_centroid_of_patch(self)

    @property
    @store_result_in_grid()
    @read_only_array
    def area_of_patch(self):
        return get_area_of_patch(self)

    @property
    @store_result_in_grid()
    @read_only_array
    def adjacent_nodes_at_node(self):
        """Get adjacent nodes.

        Examples
        --------
        >>> from landlab.graph import Graph

        First, a simple example with no diagonals.

        >>> node_x, node_y = [0, 0, 0, 1, 1, 1, 2, 2, 2], [0, 1, 2, 0, 1, 2, 0, 1, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.adjacent_nodes_at_node
        array([[ 1,  3, -1, -1],
               [ 2,  4,  0, -1],
               [ 5,  1, -1, -1],
               [ 4,  6,  0, -1],
               [ 5,  7,  3,  1],
               [ 8,  4,  2, -1],
               [ 7,  3, -1, -1],
               [ 8,  6,  4, -1],
               [ 7,  5, -1, -1]])

        Next, we add the diagonal from node 0 to node 4.

        >>> node_x, node_y = [0, 0, 0, 1, 1, 1, 2, 2, 2], [0, 1, 2, 0, 1, 2, 0, 1, 2]
        >>> links = ((0, 1), (1, 2),
        ...          (0, 3), (1, 4), (2, 5),
        ...          (3, 4), (4, 5),
        ...          (3, 6), (4, 7), (5, 8),
        ...          (6, 7), (7, 8),
        ...          (0, 4))
        >>> graph = Graph((node_y, node_x), links=links)
        >>> graph.adjacent_nodes_at_node
        array([[ 1,  4,  3, -1, -1],
               [ 2,  4,  0, -1, -1],
               [ 5,  1, -1, -1, -1],
               [ 4,  6,  0, -1, -1],
               [ 5,  7,  3,  0,  1],
               [ 8,  4,  2, -1, -1],
               [ 7,  3, -1, -1, -1],
               [ 8,  6,  4, -1, -1],
               [ 7,  5, -1, -1, -1]])
        """
        node_is_at_tail = np.choose(self.link_dirs_at_node + 1,
                                    np.array((1, -1, 0), dtype=np.int8))
        out = self.nodes_at_link[self.links_at_node, node_is_at_tail]
        out[node_is_at_tail == -1] = -1

        return out



"""Test StructuredQuadGraph."""
from numpy.testing import assert_array_equal, assert_array_almost_equal
import numpy as np

from landlab.graph import DualRadialGraph


def test_create():
    """Test creating a quad graph."""
    graph = DualRadialGraph((1, 4))

    assert graph.number_of_nodes == 5
    assert graph.number_of_links == 8
    assert graph.number_of_patches == 4

    assert graph.number_of_corners == 4
    assert graph.number_of_faces == 4
    assert graph.number_of_cells == 1


def test_spacing():
    """Test the spacing keyword for raster."""
    graph = DualRadialGraph((1, 4))

    assert_array_almost_equal(graph.xy_of_node,
                              [[ 0., -1.],
                               [-1.,  0.], [ 0.,  0.], [ 1.,  0.],
                               [ 0.,  1.]])

    assert_array_almost_equal(graph.xy_of_corner,
                              [[-0.5, -0.5], [ 0.5, -0.5],
                               [-0.5,  0.5], [ 0.5,  0.5]])


def test_spacing_keyword():
    """Test the spacing keyword for raster."""
    graph = DualRadialGraph((1, 4), spacing=2.)

    assert_array_almost_equal(graph.xy_of_node,
                              [[ 0., -2.],
                               [-2.,  0.], [ 0.,  0.], [ 2.,  0.],
                               [ 0.,  2.]])

    assert_array_almost_equal(graph.xy_of_corner,
                              [[-1., -1.], [ 1., -1.],
                               [-1.,  1.], [ 1.,  1.]])


def test_origin():
    """Test the origin keyword for raster."""
    graph = DualRadialGraph((1, 4), spacing=2., origin=(-1., 2))

    assert_array_almost_equal(graph.xy_of_node,
                              [[ 2., -3.],
                               [ 0., -1.], [ 2., -1.], [ 4., -1.],
                               [ 2.,  1.]])

    assert_array_almost_equal(graph.xy_of_corner,
                              [[ 1., -2.], [ 3., -2.],
                               [ 1.,  0.], [ 3.,  0.]])


# def test_perimeter_corners():
#     """Test the perimeter corners."""
#     y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
#     x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
#     graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
#     assert_array_equal(graph.perimeter_corners, [1, 3, 2, 0])


def test_length_of_face_and_link():
    """Test length of faces and links."""
    ROOT_2 = np.sqrt(2.)

    graph = DualRadialGraph((1, 4))

    assert_array_almost_equal(graph.length_of_link,
                              [ROOT_2, 1., ROOT_2,
                               1., 1.,
                               ROOT_2, 1., ROOT_2])
    assert_array_almost_equal(graph.length_of_face, [1., 1., 1., 1.])


def test_area_of_cell_and_patch():
    """Test areas of cells patches."""
    graph = DualRadialGraph((1, 4))

    assert_array_almost_equal(graph.area_of_patch, [.5, .5, .5, .5])
    assert_array_almost_equal(graph.area_of_cell,[1.])


def test_corners_at_cell():
    """Test corners of cells."""
    graph = DualRadialGraph((1, 4))

    assert_array_equal(graph.nodes_at_patch,
                       [[2, 1, 0], [2, 0, 3],
                        [4, 1, 2], [3, 4, 2]])
    assert_array_equal(graph.corners_at_cell, [[3, 2, 0, 1]])


def test_cells_at_corner():
    """Test areas of patches."""
    graph = DualRadialGraph((1, 4))

    assert_array_equal(graph.patches_at_node,
                       [[0, 1, -1, -1],
                        [0, 2, -1, -1], [0, 1, 2, 3], [1, 3, -1, -1],
                        [2, 3, -1, -1]])
    assert_array_equal(graph.cells_at_corner, [[0], [0], [0], [0]])


def test_cells_at_face():
    """Test cells on either side of faces."""
    graph = DualRadialGraph((1, 4))

    assert_array_equal(graph.patches_at_link,
                       [[ 0, -1], [0, 1], [ 1, -1],
                        [ 0,  2], [1, 3],
                        [ 2, -1], [2, 3], [ 3, -1]])
    assert_array_equal(graph.cells_at_face, [[0, -1], [0, -1], [0, -1], [0, -1]])


def test_faces_at_cell():
    """Test faces that form cells."""
    graph = DualRadialGraph((1, 4))
    assert_array_equal(graph.links_at_patch,
                       [[1, 3, 0], [4, 1, 2], [6, 5, 3], [4, 7, 6]])
    assert_array_equal(graph.faces_at_cell, [[2, 3, 1, 0]])


def test_corners_at_face():
    """Test corners at face tail and head."""
    graph = DualRadialGraph((1, 4))

    assert_array_equal(graph.corners_at_face,
                       [[0, 1], [0, 2], [1, 3], [2, 3]])
    assert_array_equal(graph.nodes_at_link,
                      [[1, 0], [0, 2], [0, 3],
                       [1, 2], [2, 3],
                       [1, 4], [2, 4], [3, 4]])


def test_faces_at_corner():
    """Test faces around corners."""
    graph = DualRadialGraph((1, 4))

    assert_array_equal(graph.faces_at_corner,
                       [[0, 1], [2, 0], [3, 1], [3, 2]])
    assert_array_equal(graph.links_at_node,
                      [[2, 1, 0, -1],
                       [3, 5, 0, -1], [4, 6, 3, 1], [7, 4, 2, -1],
                       [5, 6, 7, -1]])


def test_face_dirs_at_corner():
    """Test face directions at corners."""
    graph = DualRadialGraph((1, 4))

    assert_array_equal(graph.face_dirs_at_corner,
                       [[-1, -1], [-1, 1], [-1, 1], [1, 1]])
    assert_array_equal(graph.link_dirs_at_node,
                       [[-1, -1,  1, 0],
                        [-1, -1, -1, 0], [-1, -1, 1, 1], [-1, 1, 1, 0],
                        [ 1,  1,  1, 0]])



import pytest
from six import StringIO

from landlab import ModelParameterDictionary


@pytest.fixture
def pdict_setup():
    test_param_dict_file = u"""
    # A Comment
    INT_VAL:
    1

    FLOAT_VAL: # A Comment
    2.2
    STRING_VAL:
    The Landlab

    TRUE_BOOL_VAL:
    True
    FALSE_BOOL_VAL:
    False
    """

    param_file = StringIO(test_param_dict_file)
    param_dict = ModelParameterDictionary()
    param_dict.read_from_file(param_file)

    class ParamDictSetup(object):
        pass

    setup = ParamDictSetup()
    setup.param_dict = param_dict
    setup.param_dict_file = test_param_dict_file

    return setup


@pytest.fixture
def auto_type_setup():
    TEST_FILE = u"""
# A Comment
INT_VAL:
1
DBL_VAL:
1.2
STR_VAL:
landlab
BOOL_VAL:
true
INT_ARRAY_VAL:
1,2 ,4 ,7

DBL_ARRAY_VAL:
1.,2. ,4. ,7.
    """
    param_dict = ModelParameterDictionary(auto_type=True, from_file=StringIO(TEST_FILE))

    return param_dict



#! /usr/bin/env python
"""
Methods to plot data defined on Landlab grids.

Plotting functions
++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.plot.imshow.imshow_grid
    ~landlab.plot.imshow.imshow_grid_at_cell
    ~landlab.plot.imshow.imshow_grid_at_node
"""


import numpy as np
import inspect
from landlab.field.scalar_data_fields import FieldError
try:
    import matplotlib.pyplot as plt
except ImportError:
    import warnings
    warnings.warn('matplotlib not found', ImportWarning)
from landlab.grid import CLOSED_BOUNDARY
from landlab.grid.raster import RasterModelGrid
from landlab.grid.voronoi import VoronoiDelaunayGrid
from landlab.plot.event_handler import query_grid_on_button_press
from landlab.utils.decorators import deprecated


def imshow_grid_at_node(grid, values, **kwds):
    """imshow_grid_at_node(grid, values, plot_name=None, var_name=None,
                           var_units=None, grid_units=None,
                           symmetric_cbar=False, cmap='pink',
                           limits=(values.min(), values.max()),
                           vmin=values.min(), vmax=values.max(),
                           allow_colorbar=True,
                           norm=[linear], shrink=1.,
                           color_for_closed='black',
                           color_for_background=None,
                           show_elements=False, output=None)

    Prepare a map view of data over all nodes in the grid.

    Data is plotted as cells shaded with the value at the node at its center.
    Outer edges of perimeter cells are extrapolated. Closed elements are
    colored uniformly (default black, overridden with kwd 'color_for_closed');
    other open boundary nodes get their actual values.

    *values* can be a field name, a regular array, or a masked array. If a
    masked array is provided, masked entries will be treated as if they were
    Landlab CLOSED_BOUNDARYs. Used together with the color_at_closed=None
    keyword (i.e., "transparent"), this can allow for construction of overlay
    layers in a figure (e.g., only defining values in a river network, and
    overlaying it on another landscape).

    Use matplotlib functions like xlim, ylim to modify your plot after calling
    :func:`imshow_grid`, as desired.

    Node coordinates are printed when a mouse button is pressed on a cell in
    the plot.

    This function happily works with both regular and irregular grids.

    Parameters
    ----------
    grid : ModelGrid
        Grid containing the field to plot, or describing the geometry of the
        provided array.
    values : array_like, masked_array, or str
        Node values, or a field name as a string from which to draw the data.
    plot_name : str, optional
        String to put as the plot title.
    var_name : str, optional
        Variable name, to use as a colorbar label.
    var_units : str, optional
        Units for the variable being plotted, for the colorbar.
    grid_units : tuple of str, optional
        Units for y, and x dimensions. If None, component will look to the
        gri property `axis_units` for this information. If no units are
        specified there, no entry is made.
    symmetric_cbar : bool
        Make the colormap symetric about 0.
    cmap : str
        Name of a colormap
    limits : tuple of float
        Minimum and maximum of the colorbar.
    vmin, vmax: floats
        Alternatives to limits.
    allow_colorbar : bool
        If True, include the colorbar.
    colorbar_label : str or None
        The string with which to label the colorbar.
    norm : matplotlib.colors.Normalize
        The normalizing object which scales data, typically into the interval
        [0, 1]. Ignore in most cases.
    shrink : float
        Fraction by which to shrink the colorbar.
    color_for_closed : str or None
        Color to use for closed nodes (default 'black'). If None, closed
        (or masked) nodes will be transparent.
    color_for_background : color str or other color declaration, or None
        Color to use for closed elements (default None). If None, the
        background will be transparent, and appear white.
    show_elements : bool
        If True, and grid is a Voronoi, the faces will be plotted in black
        along with just the colour of the cell, defining the cell outlines
        (defaults False).
    output : None, string, or bool
        If None (or False), the image is sent to the imaging buffer to await
        an explicit call to show() or savefig() from outside this function.
        If a string, the string should be the path to a save location, and the
        filename (with file extension). The function will then call
        plt.savefig([string]) itself. If True, the function will call
        plt.show() itself once plotting is complete.
    """
    if isinstance(values, str):
        values_at_node = grid.at_node[values]
    else:
        values_at_node = values.reshape((-1, ))

    if values_at_node.size != grid.number_of_nodes:
        raise ValueError('number of values does not match number of nodes')

    values_at_node = np.ma.masked_where(
        grid.status_at_node == CLOSED_BOUNDARY, values_at_node)

    try:
        shape = grid.shape
    except AttributeError:
        shape = (-1, )

    _imshow_grid_values(grid, values_at_node.reshape(shape), **kwds)

    if isinstance(values, str):
        plt.title(values)

    plt.gcf().canvas.mpl_connect('button_press_event',
       lambda event: query_grid_on_button_press(event, grid))


@deprecated(use='imshow_grid_at_node', version='0.5')
def imshow_node_grid(grid, values, **kwds):
    imshow_grid_at_node(grid, values, **kwds)


def imshow_grid_at_cell(grid, values, **kwds):
    """imshow_grid_at_cell(grid, values, plot_name=None, var_name=None,
                           var_units=None, grid_units=None,
                           symmetric_cbar=False, cmap='pink',
                           limits=(values.min(), values.max()),
                           vmin=values.min(), vmax=values.max(),
                           allow_colorbar=True, colorbar_label=None,
                           norm=[linear], shrink=1.,
                           color_for_closed='black',
                           color_for_background=None,
                           show_elements=False, output=None)

    Map view of grid data over all grid cells.

    Prepares a map view of data over all cells in the grid.
    Method can take any of the same ``**kwds`` as :func:`imshow_grid_at_node`.

    Parameters
    ----------
    grid : ModelGrid
        Grid containing the field to plot, or describing the geometry of the
        provided array.
    values : array_like, masked_array, or str
        Values at the cells on the grid. Alternatively, can be a field name
        (string) from which to draw the data from the grid.
    plot_name : str, optional
        String to put as the plot title.
    var_name : str, optional
        Variable name, to use as a colorbar label.
    var_units : str, optional
        Units for the variable being plotted, for the colorbar.
    grid_units : tuple of str, optional
        Units for y, and x dimensions. If None, component will look to the
        gri property `axis_units` for this information. If no units are
        specified there, no entry is made.
    symmetric_cbar : bool
        Make the colormap symetric about 0.
    cmap : str
        Name of a colormap
    limits : tuple of float
        Minimum and maximum of the colorbar.
    vmin, vmax: floats
        Alternatives to limits.
    allow_colorbar : bool
        If True, include the colorbar.
    colorbar_label : str or None
        The string with which to label the colorbar.
    norm : matplotlib.colors.Normalize
        The normalizing object which scales data, typically into the interval
        [0, 1]. Ignore in most cases.
    shrink : float
        Fraction by which to shrink the colorbar.
    color_for_closed : str or None
        Color to use for closed elements (default 'black'). If None, closed
        (or masked) elements will be transparent.
    color_for_background : color str or other color declaration, or None
        Color to use for closed elements (default None). If None, the
        background will be transparent, and appear white.
    show_elements : bool
        If True, and grid is a Voronoi, the faces will be plotted in black
        along with just the colour of the cell, defining the cell outlines
        (defaults False).
    output : None, string, or bool
        If None (or False), the image is sent to the imaging buffer to await
        an explicit call to show() or savefig() from outside this function.
        If a string, the string should be the path to a save location, and the
        filename (with file extension). The function will then call
        plt.savefig([string]) itself. If True, the function will call
        plt.show() itself once plotting is complete.

    Raises
    ------
    ValueError
        If input grid is not uniform rectilinear.
    """
    if isinstance(values, str):
        try:
            values_at_cell = grid.at_cell[values]
        except FieldError:
            values_at_cell = grid.at_node[values]
    else:
        values_at_cell = values

    if values_at_cell.size == grid.number_of_nodes:
        values_at_cell = values_at_cell[grid.node_at_cell]

    if values_at_cell.size != grid.number_of_cells:
        raise ValueError('number of values must match number of cells or '
                         'number of nodes')

    values_at_cell = np.ma.asarray(values_at_cell)
    values_at_cell.mask = True
    values_at_cell.mask[grid.core_cells] = False

    myimage = _imshow_grid_values(grid,
                                  values_at_cell.reshape(grid.cell_grid_shape),
                                  **kwds)

    if isinstance(values, str):
        plt.title(values)

    return myimage


@deprecated(use='imshow_grid_at_cell', version='0.5')
def imshow_cell_grid(grid, values, **kwds):
    imshow_grid_at_cell(grid, values, **kwds)


def _imshow_grid_values(grid, values, plot_name=None, var_name=None,
                        var_units=None, grid_units=(None, None),
                        symmetric_cbar=False, cmap='pink', limits=None,
                        colorbar_label = None,
                        allow_colorbar=True, vmin=None, vmax=None,
                        norm=None, shrink=1., color_for_closed='black',
                        color_for_background=None, show_elements=False,
                        output=None):

    gridtypes = inspect.getmro(grid.__class__)

    cmap = plt.get_cmap(cmap)
    if color_for_closed is not None:
        cmap.set_bad(color=color_for_closed)
    else:
        cmap.set_bad(alpha=0.)

    if isinstance(grid, RasterModelGrid):
        if values.ndim != 2:
            raise ValueError('values must have ndim == 2')

        y = np.arange(values.shape[0] + 1) * grid.dy - grid.dy * .5
        x = np.arange(values.shape[1] + 1) * grid.dx - grid.dx * .5

        kwds = dict(cmap=cmap)
        (kwds['vmin'], kwds['vmax']) = (values.min(), values.max())
        if (limits is None) and ((vmin is None) and (vmax is None)):
            if symmetric_cbar:
                (var_min, var_max) = (values.min(), values.max())
                limit = max(abs(var_min), abs(var_max))
                (kwds['vmin'], kwds['vmax']) = (- limit, limit)
        elif limits is not None:
            (kwds['vmin'], kwds['vmax']) = (limits[0], limits[1])
        else:
            if vmin is not None:
                kwds['vmin'] = vmin
            if vmax is not None:
                kwds['vmax'] = vmax

        if np.isclose(grid.dx, grid.dy):
            if values.size == grid.number_of_nodes:
                myimage = plt.imshow(
                    values.reshape(grid.shape), origin='lower',
                    extent=(x[0], x[-1], y[0], y[-1]), **kwds)
            else:  # this is a cell grid, and has been reshaped already...
                myimage = plt.imshow(values, origin='lower',
                                     extent=(x[0], x[-1], y[0], y[-1]), **kwds)
        myimage = plt.pcolormesh(x, y, values, **kwds)

        plt.gca().set_aspect(1.)
        plt.autoscale(tight=True)

        if allow_colorbar:
            cb = plt.colorbar(norm=norm, shrink=shrink)
            if colorbar_label:
                cb.set_label(colorbar_label)
    elif VoronoiDelaunayGrid in gridtypes:
        # This is still very much ad-hoc, and needs prettifying.
        # We should save the modifications needed to plot color all the way
        # to the diagram edge *into* the grid, for faster plotting.
        # (see http://stackoverflow.com/questions/20515554/...
        # colorize-voronoi-diagram)
        # (This technique is not implemented yet)
        from scipy.spatial import voronoi_plot_2d
        import matplotlib.colors as colors
        import matplotlib.cm as cmx
        cm = plt.get_cmap(cmap)

        if (limits is None) and ((vmin is None) and (vmax is None)):
            # only want to work with NOT CLOSED nodes
            open_nodes = grid.status_at_node != 4
            if symmetric_cbar:
                (var_min, var_max) = (values.flat[
                    open_nodes].min(), values.flat[open_nodes].max())
                limit = max(abs(var_min), abs(var_max))
                (vmin, vmax) = (- limit, limit)
            else:
                (vmin, vmax) = (values.flat[
                    open_nodes].min(), values.flat[open_nodes].max())
        elif limits is not None:
            (vmin, vmax) = (limits[0], limits[1])
        else:
            open_nodes = grid.status_at_node != 4
            if vmin is None:
                vmin = values.flat[open_nodes].min()
            if vmax is None:
                vmax = values.flat[open_nodes].max()

        cNorm = colors.Normalize(vmin, vmax)
        scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=cm)
        colorVal = scalarMap.to_rgba(values)

        if show_elements:
            myimage = voronoi_plot_2d(grid.vor, show_vertices=False,
                                      show_points=False)

        # show_points to be supported in scipy0.18, but harmless for now
        mycolors = (i for i in colorVal)
        for order in grid.vor.point_region:
            region = grid.vor.regions[order]
            colortouse = next(mycolors)
            if -1 not in region:
                polygon = [grid.vor.vertices[i] for i in region]
                plt.fill(*zip(*polygon), color=colortouse)

        plt.gca().set_aspect(1.)
        # plt.autoscale(tight=True)
        # Tempting though it is to move the boundary outboard of the outermost
        # nodes (e.g., to the outermost corners), this is a bad idea, as the
        # outermost cells tend to have highly elongated shapes which make the
        # plot look stupid
        plt.xlim((np.min(grid.node_x), np.max(grid.node_x)))
        plt.ylim((np.min(grid.node_y), np.max(grid.node_y)))

        scalarMap.set_array(values)
        if allow_colorbar:
            cb = plt.colorbar(scalarMap, shrink=shrink)

    if grid_units[1] is None and grid_units[0] is None:
        grid_units = grid.axis_units
        if grid_units[1] == '-' and grid_units[0] == '-':
            plt.xlabel('X')
            plt.ylabel('Y')
        else:
            plt.xlabel('X (%s)' % grid_units[1])
            plt.ylabel('Y (%s)' % grid_units[0])
    else:
        plt.xlabel('X (%s)' % grid_units[1])
        plt.ylabel('Y (%s)' % grid_units[0])

    if plot_name is not None:
        plt.title('%s' % (plot_name))

    if var_name is not None or var_units is not None:
        if var_name is not None:
            assert type(var_name) is str
            if var_units is not None:
                assert type(var_units) is str
                colorbar_label = var_name + ' (' + var_units + ')'
            else:
                colorbar_label = var_name
        else:
            assert type(var_units) is str
            colorbar_label = '(' + var_units + ')'
        assert type(colorbar_label) is str
        assert allow_colorbar
        cb.set_label(colorbar_label)

    if color_for_background is not None:
        plt.gca().set_axis_bgcolor(color_for_background)

    if output is not None:
        if type(output) is str:
            plt.savefig(output)
            plt.clf()
        elif output:
            plt.show()


def imshow_grid(grid, values, **kwds):
    """imshow_grid(grid, values, plot_name=None, var_name=None,
                   var_units=None, grid_units=None,
                   symmetric_cbar=False, cmap='pink',
                   limits=(values.min(), values.max()),
                   vmin=values.min(), vmax=values.max(),
                   allow_colorbar=True, colorbar_label=None,
                   norm=[linear], shrink=1.,
                   color_for_closed='black',
                   color_for_background=None,
                   show_elements=False)

    Prepare a map view of data over all nodes or cells in the grid.

    Data is plotted as colored cells. If at='node', the surrounding cell is
    shaded with the value at the node at its center. If at='cell', the cell
    is shaded with its own value. Outer edges of perimeter cells are
    extrapolated. Closed elements are colored uniformly (default black,
    overridden with kwd 'color_for_closed'); other open boundary nodes get
    their actual values.

    *values* can be a field name, a regular array, or a masked array. If a
    masked array is provided, masked entries will be treated as if they were
    Landlab CLOSED_BOUNDARYs. Used together with the color_for_closed=None
    keyword (i.e., "transparent"), this can allow for construction of overlay
    layers in a figure (e.g., only defining values in a river network, and
    overlaying it on another landscape).

    Use matplotlib functions like xlim, ylim to modify your plot after calling
    :func:`imshow_grid`, as desired.

    This function happily works with both regular and irregular grids.

    Parameters
    ----------
    grid : ModelGrid
        Grid containing the field to plot, or describing the geometry of the
        provided array.
    values : array_like, masked_array, or str
        Node or cell values, or a field name as a string from which to draw
        the data.
    at : str, {'node', 'cell'}
        Tells plotter where values are defined.
    plot_name : str, optional
        String to put as the plot title.
    var_name : str, optional
        Variable name, to use as a colorbar label.
    var_units : str, optional
        Units for the variable being plotted, for the colorbar.
    grid_units : tuple of str, optional
        Units for y, and x dimensions. If None, component will look to the
        gri property `axis_units` for this information. If no units are
        specified there, no entry is made.
    symmetric_cbar : bool
        Make the colormap symetric about 0.
    cmap : str
        Name of a colormap
    limits : tuple of float
        Minimum and maximum of the colorbar.
    vmin, vmax: floats
        Alternatives to limits.
    allow_colorbar : bool
        If True, include the colorbar.
    colorbar_label : str or None
        The string with which to label the colorbar.
    norm : matplotlib.colors.Normalize
        The normalizing object which scales data, typically into the interval
        [0, 1]. Ignore in most cases.
    shrink : float
        Fraction by which to shrink the colorbar.
    color_for_closed : str or None
        Color to use for closed elements (default 'black'). If None, closed
        (or masked) elements will be transparent.
    color_for_background : color str or other color declaration, or None
        Color to use for closed elements (default None). If None, the
        background will be transparent, and appear white.
    show_elements : bool
        If True, and grid is a Voronoi, the faces will be plotted in black
        along with just the colour of the cell, defining the cell outlines
        (defaults False).
    output : None, string, or bool
        If None (or False), the image is sent to the imaging buffer to await
        an explicit call to show() or savefig() from outside this function.
        If a string, the string should be the path to a save location, and the
        filename (with file extension). The function will then call
        plt.savefig([string]) itself. If True, the function will call
        plt.show() itself once plotting is complete.
    """
    show = kwds.pop('show', False)
    values_at = kwds.pop('values_at', 'node')
    values_at = kwds.pop('at', values_at)

    if isinstance(values, str):
        values = grid.field_values(values_at, values)

    if isinstance(values, str):
        values = grid.field_values(values_at, values)

    if values_at == 'node':
        imshow_grid_at_node(grid, values, **kwds)
    elif values_at == 'cell':
        imshow_grid_at_cell(grid, values, **kwds)
    else:
        raise TypeError('value location %s not understood' % values_at)

    # retained for backwards compatibility:
    if show:
        plt.show()



import numpy as np

from ..graph import Graph
from ...utils.decorators import store_result_in_grid
from .ext.at_node import (fill_patches_at_node, fill_links_at_node,
                          fill_link_dirs_at_node, fill_perimeter_nodes)
from .ext.at_link import fill_patches_at_link, fill_nodes_at_link
from .ext.at_patch import fill_links_at_patch


def setup_horizontal_links(shape):
    from .ext.at_link import fill_horizontal_links
    n_horizontal_links = shape[0] * (shape[1] - 1)
    horizontal_links = np.empty(n_horizontal_links, dtype=int)
    fill_horizontal_links(shape, horizontal_links)
    return horizontal_links


def setup_vertical_links(shape):
    from .ext.at_link import fill_vertical_links
    n_vertical_links = (shape[0] - 1) * shape[1]
    vertical_links = np.empty(n_vertical_links, dtype=int)
    fill_vertical_links(shape, vertical_links)
    return vertical_links


def setup_perimeter_nodes(shape):
    n_perimeter_nodes = 2 * shape[0] + 2 * (shape[1] - 2)
    perimeter_nodes = np.empty(n_perimeter_nodes, dtype=int)
    fill_perimeter_nodes(shape, perimeter_nodes)
    return perimeter_nodes


def setup_link_dirs_at_node(shape):
    n_nodes = shape[0] * shape[1]
    link_dirs_at_node = np.empty((n_nodes , 4), dtype=int)
    fill_link_dirs_at_node(shape, link_dirs_at_node)
    return link_dirs_at_node


def setup_links_at_node(shape):
    n_nodes = shape[0] * shape[1]
    links_at_node = np.empty((n_nodes , 4), dtype=int)
    fill_links_at_node(shape, links_at_node)
    return links_at_node


def setup_links_at_patch(shape):
    """Get links that define patches for a raster grid.

    Examples
    --------
    >>> from landlab.graph.structured_quad.structured_quad import setup_links_at_patch
    >>> setup_links_at_patch((3, 4)) # doctest: +NORMALIZE_WHITESPACE
    array([[ 4,  7,  3,  0], [ 5,  8,  4,  1], [ 6,  9,  5,  2],
           [11, 14, 10,  7], [12, 15, 11,  8], [13, 16, 12,  9]])
    """
    n_patches = (shape[0] - 1) * (shape[1] - 1)
    links_at_patch = np.empty((n_patches , 4), dtype=int)
    fill_links_at_patch(shape, links_at_patch)
    return links_at_patch


def setup_nodes_at_link(shape):
    """
    Examples
    --------
    >>> from landlab.graph.structured_quad.structured_quad import setup_nodes_at_link
    >>> setup_nodes_at_link((3, 4)) # doctest: +NORMALIZE_WHITESPACE
    array([[ 0,  1], [ 1,  2], [ 2,  3],
           [ 0,  4], [ 1,  5], [ 2,  6], [ 3,  7],
           [ 4,  5], [ 5,  6], [ 6,  7],
           [ 4,  8], [ 5,  9], [ 6, 10], [ 7, 11],
           [ 8,  9], [ 9, 10], [10, 11]])
    """
    n_links = shape[0] * (shape[1] - 1) +  (shape[0] - 1) * shape[1]
    nodes_at_link = np.empty((n_links , 2), dtype=int)
    fill_nodes_at_link(shape, nodes_at_link)

    return nodes_at_link


def setup_patches_at_node(shape):
    n_nodes = shape[0] * shape[1]
    patches_at_node = np.empty((n_nodes , 4), dtype=int)
    fill_patches_at_node(shape, patches_at_node)

    return patches_at_node


def setup_patches_at_link(shape):
    n_links = shape[0] * (shape[1] - 1) +  (shape[0] - 1) * shape[1]
    patches_at_link = np.empty((n_links , 2), dtype=int)
    fill_patches_at_link(shape, patches_at_link)

    return patches_at_link


def setup_node_coords(shape, spacing=1., origin=0.):
    spacing = np.broadcast_to(spacing, 2)
    origin = np.broadcast_to(origin, 2)

    rows = np.arange(shape[0], dtype=float) * spacing[0] + origin[0]
    cols = np.arange(shape[1], dtype=float) * spacing[1] + origin[1]

    return setup_node_coords_rectilinear((rows, cols))


def setup_node_coords_rectilinear(coords):
    rows = np.asarray(coords[0], dtype=float)
    cols = np.asarray(coords[1], dtype=float)

    node_y_and_x = np.meshgrid(rows, cols, indexing='ij')
    return setup_node_coords_structured(node_y_and_x)


def setup_node_coords_structured(coords, shape=None):
    node_y = np.asarray(coords[0], dtype=float)
    node_x = np.asarray(coords[1], dtype=float)

    if shape:
        node_y.shape = shape
        node_x.shape = shape
    else:
        node_x.shape = node_y.shape

    if node_y.shape != node_x.shape:
        raise ValueError('shape mismatch in node x and y coordinates')

    return (node_y, node_x)


class StructuredQuadGraphExtras(object):
    @property
    def shape(self):
        return self._shape

    @property
    def number_of_node_rows(self):
        return self._shape[0]

    @property
    def number_of_node_columns(self):
        return self._shape[1]

    @property
    def nodes_at_right_edge(self):
        return np.arange(self.shape[1] - 1, self.number_of_nodes, self.shape[1])

    @property
    def nodes_at_top_edge(self):
        return np.arange(self.number_of_nodes - self.shape[1],
                         self.number_of_nodes)

    @property
    def nodes_at_left_edge(self):
        return np.arange(0, self.number_of_nodes, self.shape[1])

    @property
    def nodes_at_bottom_edge(self):
        return np.arange(self.shape[1])

    def nodes_at_edge(self, edge):
        if edge not in ('right', 'top', 'left', 'bottom'):
            raise ValueError('value for edge not understood')
        return getattr(self, 'nodes_at_{edge}_edge'.format(edge=edge))

    @property
    def horizontal_links(self):
        return setup_horizontal_links(self.shape)

    @property
    def vertical_links(self):
        return setup_vertical_links(self.shape)


class StructuredQuadGraph(StructuredQuadGraphExtras, Graph):

    """Graph of a structured grid of quadrilaterals.

    Examples
    --------
    >>> from landlab.graph import StructuredQuadGraph
    >>> node_y = [-1, -2, -3,
    ...            0,  0,  0,
    ...            1,  2,  3]
    >>> node_x = [ 0,  1,  2,
    ...            0,  2,  3,
    ...            0,  1,  2]
    >>> graph = StructuredQuadGraph((node_y, node_x), shape=(3, 3))
    >>> graph.number_of_nodes
    9
    >>> graph.y_of_node # doctest: +NORMALIZE_WHITESPACE
    array([-3., -2., -1.,
            0.,  0.,  0.,
            1.,  2.,  3.])
    >>> graph.x_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 2.,  1.,  0.,
            0.,  2.,  3.,
            0.,  1.,  2.])
    """

    def __init__(self, node_y_and_x, shape=None):
        """Create a structured grid of quadrilaterals.

        Parameters
        ----------
        nodes : tuple of array_like
            Coordinates of every node. First *y*, then *x*.
        shape : tuple, optional
            Shape of the grid. Otherwise, use the shape of coordinate arrays.
        """
        from ..ugrid import ugrid_from_structured_quad

        if shape is None:
            self._shape = node_y_and_x[0].shape
        else:
            self._shape = shape
        mesh = ugrid_from_structured_quad(node_y_and_x, shape=shape)
        Graph.__init__(self, mesh)

    @property
    # @store_result_in_grid()
    def perimeter_nodes(self):
        return setup_perimeter_nodes(self.shape)

    @property
    @store_result_in_grid()
    def links_at_node(self):
        return setup_links_at_node(self.shape)

    @property
    # @store_result_in_grid()
    def link_dirs_at_node(self):
        return setup_link_dirs_at_node(self.shape)

    @property
    @store_result_in_grid()
    def patches_at_node(self):
        return setup_patches_at_node(self.shape)

    @property
    @store_result_in_grid()
    def patches_at_link(self):
        return setup_patches_at_link(self.shape)


class RectilinearGraph(StructuredQuadGraph):

    """Graph of a rectlinear grid of nodes.

    Examples
    --------
    >>> from landlab.graph import RectilinearGraph
    >>> graph = RectilinearGraph(([0, 1, 2, 3], [1, 4, 8]))
    >>> graph.number_of_nodes
    12
    >>> graph.y_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 0., 0., 0.,
            1., 1., 1.,
            2., 2., 2.,
            3., 3., 3.])
    >>> graph.x_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 1., 4., 8.,
            1., 4., 8.,
            1., 4., 8.,
            1., 4., 8.])
    """

    def __init__(self, nodes):
        node_y_and_x = setup_node_coords_rectilinear(nodes)

        super(RectilinearGraph, self).__init__(node_y_and_x)


class UniformRectilinearGraph(StructuredQuadGraph):

    """Graph of a structured grid of quadrilaterals.

    Examples
    --------
    >>> from landlab.graph import UniformRectilinearGraph
    >>> graph = UniformRectilinearGraph((4, 3), spacing=(1, 2), origin=(-1, 0))
    >>> graph.number_of_nodes
    12
    >>> graph.y_of_node # doctest: +NORMALIZE_WHITESPACE
    array([-1., -1., -1.,
            0.,  0.,  0.,
            1.,  1.,  1.,
            2.,  2.,  2.])
    >>> graph.x_of_node # doctest: +NORMALIZE_WHITESPACE
    array([ 0.,  2.,  4.,
            0.,  2.,  4.,
            0.,  2.,  4.,
            0.,  2.,  4.])
    >>> graph.links_at_node # doctest: +NORMALIZE_WHITESPACE
    array([[ 0,  2, -1, -1], [ 1,  3,  0, -1], [-1,  4,  1, -1],
           [ 5,  7, -1,  2], [ 6,  8,  5,  3], [-1,  9,  6,  4],
           [10, 12, -1,  7], [11, 13, 10,  8], [-1, 14, 11,  9],
           [15, -1, -1, 12], [16, -1, 15, 13], [-1, -1, 16, 14]])
    >>> graph.link_dirs_at_node # doctest: +NORMALIZE_WHITESPACE
    array([[-1, -1,  0,  0], [-1, -1,  1,  0], [ 0, -1,  1,  0],
           [-1, -1,  0,  1], [-1, -1,  1,  1], [ 0, -1,  1,  1],
           [-1, -1,  0,  1], [-1, -1,  1,  1], [ 0, -1,  1,  1],
           [-1,  0,  0,  1], [-1,  0,  1,  1], [ 0,  0,  1,  1]])
    >>> graph.nodes_at_link # doctest: +NORMALIZE_WHITESPACE
    array([[ 0,  1], [ 1,  2], [ 0,  3], [ 1,  4], [ 2,  5],
           [ 3,  4], [ 4,  5], [ 3,  6], [ 4,  7], [ 5,  8],
           [ 6,  7], [ 7,  8], [ 6,  9], [ 7, 10], [ 8, 11],
           [ 9, 10], [10, 11]])
    >>> graph.links_at_patch # doctest: +NORMALIZE_WHITESPACE
    array([[ 3,  5,  2,  0], [ 4,  6,  3,  1],
           [ 8, 10,  7,  5], [ 9, 11,  8,  6],
           [13, 15, 12, 10], [14, 16, 13, 11]])
    >>> graph.nodes_at_patch # doctest: +NORMALIZE_WHITESPACE
    array([[ 4,  3,  0,  1], [ 5,  4,  1,  2],
           [ 7,  6,  3,  4], [ 8,  7,  4,  5],
           [10,  9,  6,  7], [11, 10,  7,  8]])
    """

    def __init__(self, shape, spacing=1., origin=0.):
        node_y_and_x = setup_node_coords(shape, spacing, origin)

        super(UniformRectilinearGraph, self).__init__(node_y_and_x)



#! /usr/bin/env python
from landlab import RasterModelGrid
from landlab.components import FlowAccumulator
import pickle 
from numpy.testing import assert_array_equal
import os
from landlab.io.native_landlab import save_grid, load_grid

def compare_dictionaries(dict_1, dict_2, dict_1_name, dict_2_name, path=""):
    """Compare two dictionaries recursively to find non mathcing elements

    Args:
        dict_1: dictionary 1
        dict_2: dictionary 2

    Returns:

    """
    import numpy as np
    err = ''
    key_err = ''
    value_err = ''
    old_path = path
    for k in dict_1.keys():
        path = old_path + "[%s]" % k
        if not k in dict_2:
            key_err += "Key %s%s not in %s\n" % (dict_2_name, path, dict_2_name)
        else:
            if isinstance(dict_1[k], dict) and isinstance(dict_2[k], dict):
                err += compare_dictionaries(dict_1[k],dict_2[k],'d1','d2', path)
            else:
                o1 = dict_1[k]
                o2 = dict_2[k]
                try:
                    if o1  != o2:
                        value_err += "Value of %s%s (%s) not same as %s%s (%s)\n"\
                            % (dict_1_name, path, dict_1[k], dict_2_name, path, dict_2[k])
                            
                except ValueError:
                    if np.array_equal(np.asarray(o1), np.asarray(o1)) == False:
                        value_err += "Value of %s%s (%s) not same as %s%s (%s)\n"\
                            % (dict_1_name, path, dict_1[k], dict_2_name, path, dict_2[k])
                    
    for k in dict_2.keys():
        path = old_path + "[%s]" % k
        if not k in dict_1:
            key_err += "Key %s%s not in %s\n" % (dict_2_name, path, dict_1_name)

    return key_err + value_err + err


def test_pickle():
    # Make a simple-ish grid
    mg1 = RasterModelGrid(10,10,2.)
    z = mg1.add_zeros('node', 'topographic__elevation')
    z += mg1.node_x.copy()
    fa = FlowAccumulator(mg1, flow_director='D8')
    fa.run_one_step()
    
    # save it with pickle
    with open('testsavedgrid.grid', 'wb') as f:
        pickle.dump(mg1, f)
    
    # load it with pickle
    with open('testsavedgrid.grid', 'rb') as f:
        mg2 = pickle.load(f)

    os.remove('testsavedgrid.grid')

    assert mg1.shape == mg2.shape
    assert (mg1.dy, mg1.dx) == (mg2.dy, mg2.dx)
    assert_array_equal(mg1.status_at_node, mg2.status_at_node)
    for name in mg1.at_node:
        assert_array_equal(mg1.at_node[name], mg2.at_node[name])

    # compare the two
    # try:
    #     len(mg1.__dict__) == len(mg2.__dict__)
    #     # mg1keys = sorted(list(mg1.__dict__.keys()))
    #     # mg2keys = sorted(list(mg2.__dict__.keys()))
    #     mg1keys = set(mg1.__dict__.keys())
    #     mg2keys = set(mg2.__dict__.keys())

    #     assert_equal(mg1keys - mg2keys, set())
    #     assert_equal(mg2keys - mg1keys, set())
    #     # for i in range(len(mg1keys)):
    #     #     assert_equal(mg1keys[i], mg2keys[i])

    #     a = compare_dictionaries(mg1.__dict__,mg2.__dict__,'m1','m2')
    #     assert_equal(a, '')
    # except Exception:
    #     raise
    # finally:
    #     os.remove('testsavedgrid.grid')


def test_save():
    # Make a simple-ish grid
    mg1 = RasterModelGrid(10,10,2.)
    z = mg1.add_zeros('node', 'topographic__elevation')
    z += mg1.node_x.copy()
    fa = FlowAccumulator(mg1, flow_director='D8')
    fa.run_one_step()
    
    save_grid(mg1, 'testsavedgrid.grid')
    
    mg2 = load_grid('testsavedgrid.grid')
    
    os.remove('testsavedgrid.grid')

    assert mg1.shape == mg2.shape
    assert (mg1.dy, mg1.dx) == (mg2.dy, mg2.dx)
    assert_array_equal(mg1.status_at_node, mg2.status_at_node)
    for name in mg1.at_node:
        assert_array_equal(mg1.at_node[name], mg2.at_node[name])

    # compare the two
    # try:
    #     len(mg1.__dict__) == len(mg2.__dict__)
    #     mg1keys = sorted(list(mg1.__dict__.keys()))
    #     mg2keys = sorted(list(mg2.__dict__.keys()))
        
    #     for i in range(len(mg1keys)):
    #         assert_equal(mg1keys[i], mg2keys[i])

    #     a = compare_dictionaries(mg1.__dict__,mg2.__dict__,'m1','m2')
    #     assert_equal(a, '')
    # except Exception:
    #     raise
    # finally:
    #     os.remove('testsavedgrid.grid')



from .generate_detachment_ltd_erosion import DetachmentLtdErosion
from .generate_erosion_by_depth_slope import DepthSlopeProductErosion

__all__ = ['DetachmentLtdErosion',  'DepthSlopeProductErosion']


#! /usr/bin/env

#import landlab.utils.count_repeats
#from landlab.utils.count_repeats import count_repeats
from .count_repeats import count_repeated_values
from .source_tracking_algorithm import (
    track_source,
    convert_arc_flow_directions_to_landlab_node_ids,
    find_unique_upstream_hsd_ids_and_fractions)
from .watershed import (get_watershed_mask,
                        get_watershed_masks_with_area_threshold,
                        get_watershed_nodes,
                        get_watershed_outlet)



from .flow_director_d8 import FlowDirectorD8
from .flow_director_steepest import FlowDirectorSteepest
from .flow_director_mfd import FlowDirectorMFD
from .flow_director_dinf import FlowDirectorDINF
from ..flow_director import flow_direction_DN
from ..flow_director.flow_direction_DN import flow_directions


__all__ = ['FlowDirectorD8',
           'FlowDirectorSteepest',
           'FlowDirectorMFD',
           'FlowDirectorDINF',
           'flow_directions', 'flow_direction_DN']



import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_nodes_at_face(shape, np.ndarray[DTYPE_t, ndim=2] nodes_at_face):
    """Get nodes on either side of a face.

    Parameters
    ----------
    shape : tuple of int
        Shape of the grid as `(n_rows, n_cols)`.
    nodes_at_face : ndarray of int, shape `(n_faces, 2)`
        Buffer into which to place node identifiers.
    """
    cdef int face
    cdef int node
    cdef int row
    cdef int col
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]

    # Horizontal faces first
    face = 0
    for row in range(n_rows - 1):
        for col in range(1, n_cols - 1):
            node = row * n_cols + col
            nodes_at_face[face, 0] = node
            nodes_at_face[face, 1] = node + n_cols
            face += 1
        face += n_cols - 1

    # Vertical faces next
    face = n_cols - 2
    for row in range(1, n_rows - 1):
        for col in range(n_cols - 1):
            node = row * n_cols + col
            nodes_at_face[face, 0] = node
            nodes_at_face[face, 1] = node + 1
            face += 1
        face += n_cols - 2



import numpy as np
cimport numpy as np
cimport cython

from libc.stdlib cimport malloc, free, qsort

from ...sort.ext.argsort cimport unique_int


@cython.boundscheck(True)
def get_rightmost_edge_at_patch(
    np.ndarray[long, ndim=2, mode="c"] links_at_patch,
    np.ndarray[double, ndim=2, mode="c"] xy_of_link,
    np.ndarray[long, ndim=1, mode="c"] edge):
    cdef int n_patches = links_at_patch.shape[0]
    cdef int n_cols = links_at_patch.shape[1]
    cdef int patch
    cdef int link
    cdef int n
    cdef int max_n
    cdef float max_x

    for patch in range(n_patches):
        link = links_at_patch[patch, 0]
        max_x, max_n = xy_of_link[link][0], 0

        for n in range(1, n_cols):
            link = links_at_patch[patch, n]
            if link == -1:
                break
            if xy_of_link[link][0] > max_x:
                max_x, max_n = xy_of_link[link][0], n
        edge[patch] = max_n


cdef find_common_node(long * link_a, long * link_b):
    if link_a[0] == link_b[0] or link_a[0] == link_b[1]:
        return link_a[0]
    elif link_a[1] == link_b[0] or link_a[1] == link_b[1]:
        return link_a[1]
    else:
        raise ValueError('links are not connected')


cdef all_nodes_at_patch(long * links_at_patch, long max_links,
                        long * nodes_at_link, long * out):
    cdef long n_links = max_links
    cdef long link
    cdef long i
    cdef long n_nodes = 0

    while links_at_patch[n_links - 1] == -1:
        n_links -= 1

    for i in range(n_links):
        link = links_at_patch[i]

        out[n_nodes] = nodes_at_link[link * 2]
        out[n_nodes + 1] = nodes_at_link[link * 2 + 1]

        n_nodes += 2

    return n_links


cdef order_nodes_at_patch(long * all_nodes_at_patch, long * out,
                          long n_vertices):
    cdef long i
    cdef long vertex

    out[0] = all_nodes_at_patch[1]
    for vertex in range(n_vertices - 1):
        i = vertex * 2
        while all_nodes_at_patch[i] != out[vertex]:
            i += 1
        if i % 2 == 0:
            out[vertex + 1] = all_nodes_at_patch[i + 1]
        else:
            out[vertex + 1] = all_nodes_at_patch[i - 1]


@cython.boundscheck(True)
def get_nodes_at_patch(np.ndarray[long, ndim=2, mode="c"] links_at_patch,
                       np.ndarray[long, ndim=2, mode="c"] nodes_at_link,
                       np.ndarray[long, ndim=2, mode="c"] nodes_at_patch):
    cdef int n_patches = links_at_patch.shape[0]
    cdef int max_links_at_patch = links_at_patch.shape[1]
    cdef int patch
    cdef int link
    cdef int i
    cdef int n_links
    cdef int n_unique
    cdef long * all_nodes = <long *>malloc(2 * links_at_patch.shape[1] * sizeof(long))

    try:
        for patch in range(n_patches):
            n_links = _nodes_at_patch(
                &links_at_patch[patch, 0], max_links_at_patch,
                &nodes_at_link[0, 0], &nodes_at_patch[patch, 0])
    finally:
        free(all_nodes)


cdef _nodes_at_patch(long * links_at_patch, long max_links,
                     long * nodes_at_link, long * out):
    cdef long n_links = max_links
    cdef long link, next_link, prev_link
    cdef long i

    while links_at_patch[n_links - 1] == -1:
        n_links -= 1

    next_link = links_at_patch[0]
    for i in range(0, n_links - 1):
        link, next_link = next_link, links_at_patch[i + 1]

        out[i] = find_common_node(&nodes_at_link[link * 2],
                                  &nodes_at_link[next_link * 2])

    link, next_link = links_at_patch[n_links - 1], links_at_patch[0]
    out[n_links - 1] = find_common_node(&nodes_at_link[link * 2],
                                        &nodes_at_link[next_link * 2])

    return n_links



#! /usr/bin/env
"""Exceptions to raise for the netcdf module."""


class Error(Exception):

    """Base class for errors in this package."""

    pass


class NotRasterGridError(Error):

    """Raise if grid is not uniform rectilinear.

    Raise this error if the grid defined in the netcdf file is not
    uniform rectilinear with constant spacing in all dimensions.
    """

    pass



#! /usr/bin/env python
"""Functions to work with watersheds of model grids."""

from landlab import FieldError
import numpy as np


def get_watershed_mask(grid, outlet_id):
    """
    Get the watershed of an outlet returned as a boolean array.

    Parameters
    ----------
    grid : RasterModelGrid
        A landlab RasterModelGrid.
    outlet_id : integer
        The id of the outlet node.

    Returns
    -------
    watershed_mask : boolean ndarray
        True elements of this array correspond to nodes with flow that is
        received by the outlet. The length of the array is equal to the grid
        number of nodes.

    Examples
    --------

    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowRouter
    >>> from landlab.utils import get_watershed_mask

    >>> rmg = RasterModelGrid((7, 7), 1)
    >>> z = np.array([
    ...     -9999., -9999., -9999., -9999., -9999., -9999., -9999.,
    ...     -9999.,    26.,     0.,    30.,    32.,    34., -9999.,
    ...     -9999.,    28.,     1.,    25.,    28.,    32., -9999.,
    ...     -9999.,    30.,     3.,     3.,    11.,    34., -9999.,
    ...     -9999.,    32.,    11.,    25.,    18.,    38., -9999.,
    ...     -9999.,    34.,    32.,    34.,    36.,    40., -9999.,
    ...     -9999., -9999., -9999., -9999., -9999., -9999., -9999.])
    >>> rmg.at_node['topographic__elevation'] = z

    Only the bottom boundary is set to open.
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, False)
    >>> rmg.set_fixed_value_boundaries_at_grid_edges(False, False, False, True)

    Route flow.
    >>> fr = FlowRouter(rmg)
    >>> fr.run_one_step()

    >>> get_watershed_mask(rmg, 2)
    array([False, False,  True, False, False, False, False, False, False,
            True, False, False, False, False, False,  True,  True,  True,
            True,  True, False, False,  True,  True,  True,  True,  True,
           False, False,  True,  True,  True,  True,  True, False, False,
            True,  True,  True,  True,  True, False, False, False, False,
           False, False, False, False], dtype=bool)
    """
    if 'flow__receiver_node' not in grid.at_node:
        raise FieldError("A 'flow__receiver_node' field is required at the "
                         "nodes of the input grid.")

    grid_nodes = grid.nodes.flatten()
    receiver_at_node = grid.at_node['flow__receiver_node']

    # Prepare output.
    watershed_mask = np.zeros(grid.number_of_nodes, dtype=bool)

    for node in grid_nodes:
        # Follow flow path of each node.
        receiver_node = receiver_at_node[node]
        outlet_not_found = True

        while outlet_not_found:
            node_flows_to_outlet = any([receiver_node == outlet_id, 
                                        receiver_at_node[receiver_node] ==
                                        outlet_id])
            node_is_outlet = node == outlet_id

            if node_flows_to_outlet or node_is_outlet:
                watershed_mask[node] = True
                outlet_not_found = False

            else:
                receiver_node = receiver_at_node[receiver_node]
       
                if receiver_node == receiver_at_node[receiver_node]:
                    # Receiver_node is a pit.
                    outlet_not_found = False        

    return watershed_mask


def get_watershed_nodes(grid, outlet_id):
    """
    Get the watershed of an outlet returned as a list of nodes.

    Parameters
    ----------
    grid : RasterModelGrid
        A landlab RasterModelGrid.
    outlet_id : integer
        The id of the outlet node.

    Returns
    -------
    watershed_nodes : integer ndarray
        The ids of the nodes that flow to the node with the id, outlet_id.

    Examples
    --------

    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowRouter
    >>> from landlab.utils import get_watershed_nodes

    >>> rmg = RasterModelGrid((7, 7), 1)
    >>> z = np.array([
    ...     -9999., -9999., -9999., -9999., -9999., -9999., -9999.,
    ...     -9999.,    26.,     0.,    30.,    32.,    34., -9999.,
    ...     -9999.,    28.,     1.,    25.,    28.,    32., -9999.,
    ...     -9999.,    30.,     3.,     3.,    11.,    34., -9999.,
    ...     -9999.,    32.,    11.,    25.,    18.,    38., -9999.,
    ...     -9999.,    34.,    32.,    34.,    36.,    40., -9999.,
    ...     -9999., -9999., -9999., -9999., -9999., -9999., -9999.])

    >>> rmg.at_node['topographic__elevation'] = z
    >>> rmg.set_watershed_boundary_condition_outlet_id(2, z,
    ...                                                nodata_value=-9999.)

    Route flow.
    >>> fr = FlowRouter(rmg)
    >>> fr.run_one_step()

    Get the nodes of two watersheds.
    >>> mainstem_watershed_nodes = get_watershed_nodes(rmg, 2)
    >>> tributary_watershed_nodes = get_watershed_nodes(rmg, 24)

    Given the watershed boundary conditions, the number of mainstem watershed
    nodes should be equal to the number of core nodes plus the outlet node.
    >>> len(mainstem_watershed_nodes) == rmg.number_of_core_nodes + 1
    True
    """
    ws_mask = get_watershed_mask(grid, outlet_id)
    ws_nodes = np.where(ws_mask)[0]

    return ws_nodes


def get_watershed_masks_with_area_threshold(grid, critical_area):
    """
    Get masks of all of the watersheds with a minimum drainage area size.

    Parameters
    ----------
    grid : RasterModelGrid
        A landlab RasterModelGrid.
    critical_area : integer or float
        The minimum drainage area of the watersheds to identify.

    Returns
    -------
    watershed_masks : integer ndarray
        The length of the array is equal to the grid number of nodes. Values of
        this array are the watershed ids. The value of a watershed id is the
        node id of the watershed outlet. Nodes with a value of -1 have only
        downstream nodes with drainage areas below `critical_area`.

    Examples
    --------

    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowRouter
    >>> from landlab.utils import get_watershed_masks_with_area_threshold

    Create a grid with a node spacing of 200 meter.
    >>> rmg = RasterModelGrid((7, 7), 200)
    >>> z = np.array([
    ...     -9999., -9999., -9999., -9999., -9999., -9999., -9999.,
    ...     -9999.,    26.,     0.,    26.,    30.,    34., -9999.,
    ...     -9999.,    28.,     1.,    28.,     5.,    32., -9999.,
    ...     -9999.,    30.,     3.,    30.,    10.,    34., -9999.,
    ...     -9999.,    32.,    11.,    32.,    15.,    38., -9999.,
    ...     -9999.,    34.,    32.,    34.,    36.,    40., -9999.,
    ...     -9999., -9999., -9999., -9999., -9999., -9999., -9999.])
    >>> rmg.at_node['topographic__elevation'] = z
    >>> rmg.set_closed_boundaries_at_grid_edges(True, True, True, False)

    Route flow.
    >>> fr = FlowRouter(rmg)
    >>> fr.run_one_step()

    Get the masks of watersheds greater than or equal to 80,000 square-meters.
    >>> critical_area = 80000
    >>> mask = get_watershed_masks_with_area_threshold(rmg, critical_area)

    # Verify that all mask null nodes have a drainage area below critical area.
    >>> null_nodes = np.where(mask == -1)[0]
    >>> A = rmg.at_node['drainage_area'][null_nodes]
    >>> below_critical_area_nodes = A < critical_area
    >>> np.all(below_critical_area_nodes)
    True
    """

    nodes = grid.nodes.flatten()
    core_nodes = grid.node_is_core(nodes)
    boundary_nodes = grid.node_is_boundary(nodes)

    # Find outlets along grid boundary.
    A = grid.at_node['drainage_area']
    stream_nodes = A >= critical_area
    outlet_nodes = np.all([stream_nodes, boundary_nodes], 0)
    outlet_nodes = np.where(outlet_nodes)[0]

    # Get watersheds of boundary outlets.
    watershed_masks = np.repeat(-1, grid.number_of_nodes)
    for outlet in outlet_nodes:
        _assign_outlet_id_to_watershed(grid, outlet, watershed_masks)

    # Get internal watersheds.
    node_is_unresolved = np.all([A >= critical_area, watershed_masks == -1,
                                 core_nodes], 0)
    while np.any(node_is_unresolved):
        outlet = np.all([A == max(A[node_is_unresolved]), node_is_unresolved],
                         0)

        # Get the outlet with the greatest drainage area of the unresolved
        # nodes.
        outlet = np.where(outlet)[0][0]
        _assign_outlet_id_to_watershed(grid, outlet, watershed_masks)
        node_is_unresolved = np.all([A >= critical_area, watershed_masks == -1,
                                     core_nodes], 0)

    return watershed_masks


def _assign_outlet_id_to_watershed(grid, outlet, masks):
    ws_mask = get_watershed_mask(grid, outlet)
    ws_number_of_nodes = len(np.where(ws_mask)[0])
    masks[ws_mask] = np.repeat(outlet, ws_number_of_nodes)


def get_watershed_outlet(grid, source_node_id):
    """
    Get the downstream-most node (the outlet) of the source node.

    Parameters
    ----------
    grid : RasterModelGrid
        A landlab RasterModelGrid.
    source_node_id : integer
        The id of the node in which to identify its outlet.

    Returns
    -------
    outlet_node : integer
        The id of the node that is the downstream-most node (the outlet) of the
        source node.

    Examples
    --------

    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab.components import FlowRouter
    >>> from landlab.utils import get_watershed_outlet

    >>> rmg = RasterModelGrid((7, 7), 1)
    >>> z = np.array([
    ...     -9999., -9999., -9999., -9999., -9999., -9999., -9999.,
    ...     -9999.,    26.,     0.,    30.,    32.,    34., -9999.,
    ...     -9999.,    28.,     1.,    25.,    28.,    32., -9999.,
    ...     -9999.,    30.,     3.,     3.,    11.,    34., -9999.,
    ...     -9999.,    32.,    11.,    25.,    18.,    38., -9999.,
    ...     -9999.,    34.,    32.,    34.,    36.,    40., -9999.,
    ...     -9999., -9999., -9999., -9999., -9999., -9999., -9999.])

    >>> rmg.at_node['topographic__elevation'] = z
    >>> imposed_outlet = 2
    >>> rmg.set_watershed_boundary_condition_outlet_id(imposed_outlet, z,
    ...                                                nodata_value=-9999.)

    Route flow.
    >>> fr = FlowRouter(rmg)
    >>> fr.run_one_step()

    Get the grid watershed outlet.
    >>> determined_outlet = get_watershed_outlet(rmg, 40)
    >>> determined_outlet == imposed_outlet
    True
    """
    if 'flow__receiver_node' not in grid.at_node:
        raise FieldError("A 'flow__receiver_node' field is required at the "
                         "nodes of the input grid.")

    receiver_at_node = grid.at_node['flow__receiver_node']
    receiver_node = receiver_at_node[source_node_id]
    outlet_not_found = True

    while outlet_not_found:
        node_is_outlet = receiver_node == source_node_id
        node_is_boundary = grid.node_is_boundary(receiver_node)
        node_is_pit = receiver_node == receiver_at_node[receiver_node]

        if node_is_outlet or node_is_boundary or node_is_pit:
            outlet_not_found = False
            outlet_node = receiver_node
        else:
            receiver_node = receiver_at_node[receiver_node]

    return outlet_node



import numpy as np
cimport numpy as np
cimport cython


DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def _neighbors_at_link(np.ndarray[DTYPE_t, ndim=1] links, shape,
                       np.ndarray[DTYPE_t, ndim=2] out):
  cdef int stride
  cdef int n_links
  cdef int link
  cdef int i
  cdef is_top, is_bottom, is_left, is_right

  stride = 2 * shape[1] - 1
  n_links = (shape[0] - 1) * shape[1] + shape[0] * (shape[1] - 1)

  for i in range(links.size):
    link = links[i]

    is_top = link > (n_links - stride)
    is_bottom = link < stride
    is_left = link % stride == 0 or (link + shape[1]) % stride == 0
    is_right = (link - (shape[1] - 2)) % stride == 0 or (link + 1) % stride == 0

    if not is_right:
      out[i, 0] = link + 1

    if not is_top:
      out[i, 1] = link + stride

    if not is_left:
      out[i, 2] = link - 1

    if not is_bottom:
      out[i, 3] = link - stride



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  3 10:39:32 2017

@author: gtucker
"""
import pytest

from landlab import RasterModelGrid
from landlab.components import (DepthDependentTaylorDiffuser,
                                ExponentialWeatherer)
import numpy as np
from numpy.testing import assert_array_equal

import warnings

def test_4x7_grid_vs_analytical_solution():
    """Test against known analytical solution."""

    # Create a 4-row by 7-column grid with 10 m spacing
    mg = RasterModelGrid((4, 7), 10.0)

    # Close off top and bottom (N and S) boundaries so it becomes a 1D problem
    mg.set_closed_boundaries_at_grid_edges(False, True, False, True)

    # Create an elevation field, initially zero
    z = mg.add_zeros('node', 'topographic__elevation')

    # Instantiate components, and set their parameters. Note that traditional
    # diffusivity, D, is D = SCE x H*, where SCE is soil-creep efficiency.
    # Here we want D = 0.01 m2/yr and H* = 0,.5 m, so cwe set SCE = 0.02.
    diffuser = DepthDependentTaylorDiffuser(mg,
                                            linear_diffusivity=0.01,
                                            slope_crit=0.8,
                                            soil_transport_decay_depth=0.5)
    weatherer = ExponentialWeatherer(mg,
                                     soil_production__maximum_rate=0.0002,
                                     soil_production__decay_depth=0.5)

    # Get a reference to bedrock elevation field
    z_bedrock = mg.at_node['bedrock__elevation']

    # Estimate a reasonable time step. Here we take advantage of the fact that
    # we know the final slope at the outer links will be about 1.33. Stability
    # for the cubic term involves an "effective D" parameter, Deff, that should
    # be Deff = D (S / Sc)^2. (see notebook calcs)
    baselevel_rate = 0.0001
    dt = 250.0

    # Run for 750 ky
    for i in range(3000):

        z[mg.core_nodes] += baselevel_rate * dt
        z_bedrock[mg.core_nodes] += baselevel_rate * dt

        weatherer.calc_soil_prod_rate()
        diffuser.run_one_step(dt)

    # Test: these numbers represent equilibrium. See Jupyter notebook for
    # calculations.
    my_nodes = mg.nodes[2, :]
    assert_array_equal(np.round(z[my_nodes], 1),
                       np.array([0.0, 4.0, 6.7, 7.7, 6.7, 4.0, 0.0]))
    assert_array_equal(np.round(mg.at_node['soil__depth'][8:13], 2),
                       np.array([0.35, 0.35, 0.35, 0.35, 0.35]))

def test_raise_stability_error():
    mg = RasterModelGrid((5, 5))
    soilTh = mg.add_zeros('node', 'soil__depth')
    z = mg.add_zeros('node', 'topographic__elevation')
    BRz = mg.add_zeros('node', 'bedrock__elevation')
    z += mg.node_x.copy()**2
    BRz = z.copy() - 1.0
    soilTh[:] = z - BRz
    expweath = ExponentialWeatherer(mg)
    DDdiff = DepthDependentTaylorDiffuser(mg)
    expweath.calc_soil_prod_rate()
    with pytest.raises(RuntimeError):
        DDdiff.soilflux(10, if_unstable='raise')

def test_raise_kwargs_error():
    mg = RasterModelGrid((5, 5))
    soilTh = mg.add_zeros('node', 'soil__depth')
    z = mg.add_zeros('node', 'topographic__elevation')
    BRz = mg.add_zeros('node', 'bedrock__elevation')
    z += mg.node_x.copy()**2
    BRz = z.copy() - 1.0
    soilTh[:] = z - BRz
    with pytest.raises(TypeError):
        DepthDependentTaylorDiffuser(mg, diffusivity=1)

def test_infinite_taylor_error():

    mg = RasterModelGrid((5, 5))
    soilTh = mg.add_zeros('node', 'soil__depth')
    z = mg.add_zeros('node', 'topographic__elevation')
    BRz = mg.add_zeros('node', 'bedrock__elevation')
    z += mg.node_x.copy()**4
    BRz = z.copy() - 1.0
    soilTh[:] = z - BRz
    expweath = ExponentialWeatherer(mg)
    DDdiff = DepthDependentTaylorDiffuser(mg, nterms=400)
    expweath.calc_soil_prod_rate()
    with pytest.raises(RuntimeError):
        DDdiff.soilflux(10)

#def test_warn():
#    mg = RasterModelGrid((5, 5))
#    soilTh = mg.add_zeros('node', 'soil__depth')
#    z = mg.add_zeros('node', 'topographic__elevation')
#    BRz = mg.add_zeros('node', 'bedrock__elevation')
#    z += mg.node_x.copy()**2
#    BRz = z.copy() - 1.0
#    soilTh[:] = z - BRz
#    expweath = ExponentialWeatherer(mg)
#    DDdiff = DepthDependentTaylorDiffuser(mg)
#    expweath.calc_soil_prod_rate()
#
#    with warnings.catch_warnings(record=True) as w:
#    # Cause all warnings to always be triggered.
#        warnings.simplefilter("always")
#        # Trigger a warning.
#        DDdiff.soilflux(dt=10, if_unstable='warn')
#        # Verify some things
#        assert len(w) == 1
#        assert issubclass(w[-1].category, RuntimeWarning)

if __name__ == '__main__':
    test_4x7_grid_vs_analytical_solution()



import numpy as np

from ..dual import DualGraph
from .structured_quad import (StructuredQuadGraph, RectilinearGraph,
                              UniformRectilinearGraph, )


class DualStructuredQuadGraphExtras(object):
    @property
    def corners_at_right_edge(self):
        return self.dual.nodes_at_right_edge

    @property
    def corners_at_top_edge(self):
        return self.dual.nodes_at_top_edge

    @property
    def corners_at_left_edge(self):
        return self.dual.nodes_at_left_edge

    @property
    def corners_at_bottom_edge(self):
        return self.dual.nodes_at_bottom_edge

    @property
    def perimeter_corners(self):
        return self.dual.perimeter_nodes

    @property
    def horizontal_faces(self):
        return self.dual.horizontal_links

    @property
    def vertical_faces(self):
        return self.dual.vertical_links


class DualStructuredQuadGraph(DualStructuredQuadGraphExtras,
                              StructuredQuadGraph, DualGraph):

    """Dual graph of a structured grid of quadrilaterals.

    Examples
    --------
    >>> from landlab.graph import DualStructuredQuadGraph
    >>> node_y = [-1, -2, -3,
    ...            0,  0,  0,
    ...            1,  2,  3]
    >>> node_x = [ 0,  1,  2,
    ...            0,  2,  3,
    ...            0,  1,  2]
    >>> graph = DualStructuredQuadGraph((node_y, node_x), shape=(3, 3))
    >>> graph.number_of_corners == 4
    True
    >>> graph.y_of_corner
    array([-1.25, -0.75,  0.75,  1.25])
    >>> graph.x_of_corner
    array([ 2.  ,  0.75,  0.75,  2.  ])
    >>> graph.node_at_cell
    array([4])
    """

    def __init__(self, node_y_and_x, shape=None):
        dual_y, dual_x = get_corners(node_y_and_x, shape)
        dual_shape = dual_y.shape
        node_at_cell = get_node_at_cell(shape)
        nodes_at_face = get_nodes_at_face(shape)

        self._dual = StructuredQuadGraph((dual_y, dual_x), shape=dual_shape)
        StructuredQuadGraph.__init__(self, node_y_and_x, shape=shape)
        DualGraph.__init__(self, node_at_cell=node_at_cell,
                           nodes_at_face=nodes_at_face)


class DualRectilinearGraph(DualStructuredQuadGraph):

    """Create a dual graph for a rectilinear grid.

    Examples
    --------
    >>> from landlab.graph import DualRectilinearGraph
    >>> graph = DualRectilinearGraph(([0, 1, 3], [0, 5, 15, 30]))
    >>> graph.x_of_corner # doctest: +NORMALIZE_WHITESPACE
    array([ 2.5, 10. , 22.5,
            2.5, 10. , 22.5])
    >>> graph.y_of_corner # doctest: +NORMALIZE_WHITESPACE
    array([ 0.5, 0.5, 0.5,
            2. , 2. , 2. ])
    >>> graph.number_of_cells == 2
    True
    >>> graph.faces_at_cell
    array([[3, 5, 2, 0],
           [4, 6, 3, 1]])
    """

    def __init__(self, node_y_and_x):
        shape = (len(node_y_and_x[0]), len(node_y_and_x[1]))
        node_y_and_x = np.meshgrid(*node_y_and_x, indexing='ij')

        super(DualRectilinearGraph, self).__init__(node_y_and_x, shape)


class DualUniformRectilinearGraph(DualRectilinearGraph):

    """Create a dual graph for a uniform rectilinear grid.

    Examples
    --------
    >>> from landlab.graph import DualUniformRectilinearGraph
    >>> graph = DualUniformRectilinearGraph((4, 3))
    >>> graph.x_of_corner # doctest: +NORMALIZE_WHITESPACE
    array([ 0.5, 1.5,
            0.5, 1.5,
            0.5, 1.5])
    >>> graph.y_of_corner # doctest: +NORMALIZE_WHITESPACE
    array([ 0.5, 0.5,
            1.5, 1.5,
            2.5, 2.5])
    >>> graph.number_of_cells == 2
    True
    >>> graph.faces_at_cell
    array([[2, 3, 1, 0],
           [5, 6, 4, 3]])
    """

    def __init__(self, shape, spacing=(1., 1.), origin=(0., 0.)):
        spacing = np.broadcast_to(spacing, 2)
        origin = np.broadcast_to(origin, 2)

        node_y_and_x = (np.arange(shape[0]) * spacing[0] + origin[0],
                       np.arange(shape[1]) * spacing[1] + origin[1])

        super(DualUniformRectilinearGraph, self).__init__(node_y_and_x)


def get_node_at_cell(shape):
    """Set up an array that gives the node at each cell.

    Examples
    --------
    >>> from landlab.graph.structured_quad.dual_structured_quad import get_node_at_cell
    >>> get_node_at_cell((5, 6)) # doctest: +NORMALIZE_WHITESPACE
    array([ 7,  8,  9, 10,
           13, 14, 15, 16,
           19, 20, 21, 22])
    """
    from .ext.at_cell import fill_node_at_cell

    node_at_cell = np.empty((shape[0] - 2) * (shape[1] - 2), dtype=int)

    fill_node_at_cell(shape, node_at_cell)

    return node_at_cell


def get_nodes_at_face(shape):
    """Set up an array that gives the nodes on either side of each face.

    Examples
    --------
    >>> from landlab.graph.structured_quad.dual_structured_quad import get_nodes_at_face
    >>> get_nodes_at_face((3, 4)) # doctest: +NORMALIZE_WHITESPACE
    array([[ 1,  5], [ 2,  6],
           [ 4,  5], [ 5,  6], [ 6,  7],
           [ 5,  9], [ 6, 10]])
    """
    from .ext.at_face import fill_nodes_at_face

    n_faces = (shape[1] - 2) * (shape[0] - 1) + (shape[0] - 2) * (shape[1] - 1)
    nodes_at_face = np.empty((n_faces, 2), dtype=int)
    fill_nodes_at_face(shape, nodes_at_face)

    return nodes_at_face


def get_corners(node_y_and_x, shape):
    y_of_node, x_of_node = (np.asarray(node_y_and_x[0], dtype=float),
                            np.asarray(node_y_and_x[1], dtype=float))
    y_of_node.shape = x_of_node.shape = shape
    
    x_of_corner = (x_of_node[:-1, :-1] + x_of_node[:-1, 1:] +
                   x_of_node[1:, :-1] + x_of_node[1:, 1:]) * .25
    y_of_corner = (y_of_node[:-1, :-1] + y_of_node[:-1, 1:] +
                   y_of_node[1:, :-1] + y_of_node[1:, 1:]) * .25

    return y_of_corner, x_of_corner


def reshape_nodes(y_and_x, shape=None):
    y_of_node, x_of_node = (np.asarray(y_and_x[0], dtype=float),
                            np.asarray(y_and_x[1], dtype=float))

    if y_of_node.size != x_of_node.size:
        raise ValueError('size mismatch for size of x and y')

    if shape is None:
        if y_of_node.shape != x_of_node.shape:
            raise ValueError('shape mismatch for size of x and y')
    else:
        y_of_node.shape = shape
        x_of_node.shape = shape
    return y_of_node, x_of_node



# -*- coding: utf-8 -*-
"""
test_celllab_cts.py: Unit-test function for CellLabCTS.

Created on Thu Jul  9 08:20:06 2015

@author: gtucker
"""

from numpy.testing import assert_array_equal
from landlab import RasterModelGrid, HexModelGrid
from landlab.ca.celllab_cts import Transition #X, Event
from landlab.ca.raster_cts import RasterCTS
from landlab.ca.oriented_raster_cts import OrientedRasterCTS
from landlab.ca.hex_cts import HexCTS
from landlab.ca.oriented_hex_cts import OrientedHexCTS
from heapq import heappush
from heapq import heappop
import numpy as np

# For dev
from landlab.ca.celllab_cts import _RUN_NEW


def callback_function(ca, node1, node2, time_now):
    """
    This function is passed as an argument to a transition event, and then
    called automatically by CellLabCTSModel's do_event() method.
    """
    pass
#    elapsed_time = time_now - ca.last_update_time[node1]
#    if ca.node_state[node1]==1:
#        ca.prop_data[ca.propid[node1]]+=100*elapsed_time
#    elapsed_time = time_now - ca.last_update_time[node2]
#    if ca.node_state[node2]==1:
#        ca.prop_data[ca.propid[node2]]+=100*elapsed_time
#



def test_transition():
    """Test instantiation of Transition() object."""
    t = Transition((0, 0, 0), (1, 1, 0), 1.0, name='test',
                   swap_properties=False, prop_update_fn=None)
    assert t.from_state == (0,0,0)
    assert t.to_state == (1,1,0)
    assert t.rate == 1.0
    assert t.name == 'test'
    assert t.swap_properties == False
    assert t.prop_update_fn == None


def test_raster_cts():
    """
    Tests instantiation of a RasterCTS and implementation of one transition,
    with a callback function.
    """

    # Set up a small grid with no events scheduled
    mg = RasterModelGrid(4, 4, 1.0)
    mg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    node_state_grid = mg.add_ones('node', 'node_state_map', dtype=int)
    node_state_grid[6] = 0
    ns_dict = { 0 : 'black', 1 : 'white' }
    xn_list = []
    xn_list.append( Transition((1,0,0), (0,1,0), 0.1, '', True, callback_function))
    pd = mg.add_zeros('node', 'property_data', dtype=int)
    pd[5] = 50
    ca = RasterCTS(mg, ns_dict, xn_list, node_state_grid, prop_data=pd,
                   prop_reset_value=0)

    # Test the data structures
    assert (ca.num_link_states==4), 'wrong number of link states'
    assert (ca.prop_data[5]==50), 'error in property data'
    assert (ca.num_node_states==2), 'error in num_node_states'
    assert (ca.link_orientation[-1]==0), 'error in link orientation array'
    assert (ca.link_state_dict[(1, 0, 0)]==2), 'error in link state dict'
    assert (ca.n_xn[2]==1), 'error in n_xn'
    assert (ca.node_pair[1]==(0, 1, 0)), 'error in cell_pair list'

    if _RUN_NEW:
        assert (len(ca.priority_queue._queue)==1), 'event queue has wrong size'
        assert (ca.next_trn_id.size==24), 'wrong size next_trn_id'
        assert (ca.trn_id.shape==(4, 1)), 'wrong size for xn_to'
        assert (ca.trn_id[2][0]==0), 'wrong value in xn_to'
        assert (ca.trn_to[0] == 1), 'wrong trn_to state'
        assert (ca.trn_rate[0] == 0.1), 'wrong trn rate'
        assert (ca.trn_propswap[0] == 1), 'wrong trn propswap'
        assert (ca.trn_prop_update_fn == callback_function), 'wrong prop upd'
    else:
        assert (len(ca.event_queue)==1), 'event queue has wrong size'
        assert (ca.xn_to.size==4), 'wrong size for xn_to'
        assert (ca.xn_to.shape==(4, 1)), 'wrong size for xn_to'
        assert (ca.xn_to[2][0]==1), 'wrong value in xn_to'
        assert (ca.xn_rate[2][0]==0.1), 'error in transition rate array'

    # Manipulate the data in the event queue for testing:

    if _RUN_NEW:
        # pop the scheduled event off the queue
        (event_time, index, event_link) = ca.priority_queue.pop()        
        assert (ca.priority_queue._queue==[]), \
                                'event queue should now be empty but is not'

        # engineer an event
        ca.priority_queue.push(8, 1.0)
        ca.next_update[8] = 1.0
        ca.next_trn_id[8] = 0

    else:
        # pop the scheduled event off the queue
        ev = heappop(ca.event_queue)
        assert (ca.event_queue==[]), 'event queue should now be empty but is not'
    
        # engineer an event
        ev.time = 1.0
        ev.link = 8
        ev.xn_to = 1
        ev.propswap = True
        ev.prop_update_fn = callback_function
        ca.next_update[8] = 1.0
    
        # push it onto the event queue
        heappush(ca.event_queue, ev)

    # run the CA
    ca.run(2.0)

    # some more tests.
    # Is current time advancing correctly? (should only go to 1.0, not 2.0)
    # Did the two nodes (5 and 6) correctly exchange states?
    # Did the property ID and data arrays get updated? Note that the "propswap"
    # should switch propids between nodes 5 and 6, and the callback function
    # should increase the value of prop_data in the "swap" node from 50 to 150.
    assert (ca.current_time==1.0), 'current time incorrect'
    assert (ca.node_state[5]==0), 'error in node state 5'
    assert (ca.node_state[6]==1), 'error in node state 6'
    #assert (ca.prop_data[ca.propid[6]]==150), 'error in prop swap'


def test_oriented_raster_cts():
    """Tests instantiation of an OrientedRasterCTS() object"""
    mg = RasterModelGrid(3, 3, 1.0)
    nsd = {0 : 'oui', 1 : 'non'}
    xnlist = []
    xnlist.append(Transition((0,1,0), (1,1,0), 1.0, 'hopping'))
    nsg = mg.add_zeros('node', 'node_state_grid')
    orcts = OrientedRasterCTS(mg, nsd, xnlist, nsg)

    assert orcts.num_link_states == 8
    #assert_array_equal(orcts.link_orientation, [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    assert_array_equal(orcts.link_orientation, [0, 0, 1, 1, 1, 0, 0, 1, 1, 1, 0, 0])


def test_hex_cts():
    """Tests instantiation of a HexCTS() object"""
    mg = HexModelGrid(3, 2, 1.0, orientation='vertical', reorient_links=True)
    nsd = {0 : 'zero', 1 : 'one'}
    xnlist = []
    xnlist.append(Transition((0,1,0), (1,1,0), 1.0, 'transitioning'))
    nsg = mg.add_zeros('node', 'node_state_grid')
    hcts = HexCTS(mg, nsd, xnlist, nsg)

    assert hcts.num_link_states == 4
    assert_array_equal(hcts.link_orientation, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])


def test_oriented_hex_cts():
    """Tests instantiation of an OrientedHexCTS() object"""
    mg = HexModelGrid(3, 2, 1.0, orientation='vertical', reorient_links=True)
    nsd = {0 : 'zero', 1 : 'one'}
    xnlist = []
    xnlist.append(Transition((0,1,0), (1,1,0), 1.0, 'transitioning'))
    nsg = mg.add_zeros('node', 'node_state_grid')
    ohcts = OrientedHexCTS(mg, nsd, xnlist, nsg)
    
    assert ohcts.num_link_states == 12
    #assert_array_equal(ohcts.link_orientation, [2, 1, 0, 0, 0, 2, 1, 0, 2, 1, 0])
    assert_array_equal(ohcts.link_orientation, [2, 0, 1, 0, 2, 0, 1, 0, 2, 0, 1])
    
    
def test_priority_queue():
    """Test import and use of priority queue."""
    from ..cfuncs import PriorityQueue
    
    # Create a priority queue
    pq = PriorityQueue()
    
    # push a bunch of events
    pq.push(2, 2.2)
    pq.push(5, 5.5)
    pq.push(0, 0.11)
    pq.push(4, 4.4)
    pq.push(1, 1.1)
    pq.push(3, 3.3)

    # pop a bunch of events
    (priority, index, item) = pq.pop()
    assert (priority == 0.11), 'incorrect priority in PQ test'
    assert (index == 2), 'incorrect index in PQ test'
    assert (item == 0), 'incorrect item in PQ test'

    (priority, index, item) = pq.pop()
    assert (priority == 1.1), 'incorrect priority in PQ test'
    assert (index == 4), 'incorrect index in PQ test'
    assert (item == 1), 'incorrect item in PQ test'

    (priority, index, item) = pq.pop()
    assert (priority == 2.2), 'incorrect priority in PQ test'
    assert (index == 0), 'incorrect index in PQ test'
    assert (item == 2), 'incorrect item in PQ test'

    (priority, index, item) = pq.pop()
    assert (priority == 3.3), 'incorrect priority in PQ test'
    assert (index == 5), 'incorrect index in PQ test'
    assert (item == 3), 'incorrect item in PQ test'

    (priority, index, item) = pq.pop()
    assert (priority == 4.4), 'incorrect priority in PQ test'
    assert (index == 3), 'incorrect index in PQ test'
    assert (item == 4), 'incorrect item in PQ test'

    (priority, index, item) = pq.pop()
    assert (priority == 5.5), 'incorrect priority in PQ test'
    assert (index == 1), 'incorrect index in PQ test'
    assert (item == 5), 'incorrect item in PQ test'

def test_run_oriented_raster():
    """Test running with a small grid, 2 states, 4 transition types."""

    # Create an OrientedRaster with a 3x5 raster grid. Test model has 2 node
    # states and 4 transition types.
    grid = RasterModelGrid((3, 5))
    nsd = {0 : 'zero', 1 : 'one'}
    trn_list = []
    trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
    trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
    trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
    trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
    ins = np.arange(15) % 2  # makes a checkerboard pattern
    cts = OrientedRasterCTS(grid, nsd, trn_list, ins)

    # Run to 1st transition, at ~0.12
    cts.run(0.15)
    assert_array_equal(cts.node_state,
                       [0, 1, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 0])

    # Run to 2nd transition, at ~0.19
    cts.run(0.2)
    assert_array_equal(cts.node_state,
                       [0, 1, 0, 1, 0, 1, 1, 1, 1, 1, 0, 1, 0, 1, 0])

    # Run to 3rd transition, at ~0.265
    cts.run(0.27)
    assert_array_equal(cts.node_state,
                       [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0])

    # Run to 4th transition, at ~0.276 (transition is ignored)
    cts.run(0.28)
    assert_array_equal(cts.node_state,
                       [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0])

    # Run to 5th transition, at ~0.461 (ignored)
    cts.run(0.5)
    assert_array_equal(cts.node_state,
                       [0, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 1, 0])

    # Run to 6th transition, at ~0.648
    cts.run(0.65)
    assert_array_equal(cts.node_state,
                       [0, 1, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 0])


def test_grain_hill_model():
    """Run a lattice-grain-based hillslope evolution model."""
    from .grain_hill import GrainHill

    params = {
        'number_of_node_rows' : 10,
        'number_of_node_columns' : 10,
        'report_interval' : 5.0,
        'run_duration' : 10.0,
        'output_interval' : 1.0e5,
        'settling_rate' : 220000000.0,
        'disturbance_rate' : 0.01,
        'uplift_interval' : 4000.0,
        'friction_coef' : 1.0,
        'plot_interval' : 1.0,
        'show_plots' : False,
        }
    grid_size = (int(params['number_of_node_rows']), 
                 int(params['number_of_node_columns']))
    grain_hill_model = GrainHill(grid_size, **params)
    grain_hill_model.run()

    # Now test
    assert_array_equal(grain_hill_model.grid.at_node['node_state'][:18],
                       [8, 7, 7, 7, 7, 7, 7, 7, 7, 8, 0, 7, 7, 7, 7, 0, 7, 7])

    # Try with an uplift step
    params['uplift_interval'] = 5.0
    params['run_duration'] = 15.0
    grain_hill_model = GrainHill(grid_size, **params)
    grain_hill_model.run()

    # Test
    assert_array_equal(grain_hill_model.grid.at_node['node_state'][20:38],
                       [0, 7, 7, 7, 7, 0, 7, 7, 7, 0, 0, 0, 7, 7, 0, 0, 0, 7])


def test_setup_transition_data():
    """Test the CellLabCTSModel setup_transition_data method."""
    grid = RasterModelGrid((3, 4))
    nsd = {0 : 'zero', 1 : 'one'}
    trn_list = []
    trn_list.append(Transition((0, 1, 0), (1, 0, 0), 1.0))
    trn_list.append(Transition((1, 0, 0), (0, 1, 0), 2.0))
    trn_list.append(Transition((0, 1, 1), (1, 0, 1), 3.0))
    trn_list.append(Transition((0, 1, 1), (1, 1, 1), 4.0))
    ins = np.arange(12) % 2
    cts = OrientedRasterCTS(grid, nsd, trn_list, ins)

    assert_array_equal(cts.n_trn, [0, 1, 1, 0, 0, 2, 0, 0])
    assert_array_equal(cts.trn_id,
                       [[0, 0],
                        [0, 0],
                        [1, 0],
                        [0, 0],
                        [0, 0],
                        [2, 3],
                        [0, 0],
                        [0, 0]])
    assert_array_equal(cts.trn_to, [2, 1, 6, 7])
    assert_array_equal(cts.trn_rate, [1., 2., 3., 4.])


if __name__ == '__main__':
    test_transition()
    test_raster_cts()
    test_oriented_raster_cts()
    test_hex_cts()
    test_oriented_hex_cts()
    test_run_oriented_raster()
    test_grain_hill_model()
